# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import unittest
from unittest import mock
import os

from azure.core.exceptions import HttpResponseError

from azure.mgmt.web import WebSiteManagementClient
from knack.util import CLIError
from azure.cli.core.azclierror import (InvalidArgumentValueError,
                                       MutuallyExclusiveArgumentError,
                                       ArgumentUsageError,
                                       AzureResponseError)
from azure.cli.command_modules.appservice.custom import (set_deployment_user,
                                                         update_git_token, add_hostname,
                                                         update_site_configs,
                                                         get_external_ip,
                                                         view_in_browser,
                                                         sync_site_repo,
                                                         _match_host_names_from_cert,
                                                         bind_ssl_cert,
                                                         list_publish_profiles,
                                                         show_app,
                                                         get_streaming_log,
                                                         download_historical_logs,
                                                         validate_container_app_create_options,
                                                         restore_deleted_webapp,
                                                         list_snapshots,
                                                         restore_snapshot,
                                                         create_managed_ssl_cert,
                                                         add_github_actions,
                                                         update_app_settings,
                                                         update_application_settings_polling,
                                                         update_webapp,
                                                         list_startup_logs,
                                                         show_startup_log,
                                                         create_webapp)

# pylint: disable=line-too-long
from azure.cli.core.profiles import ResourceType


def _get_test_cmd():
    from azure.cli.core.mock import DummyCli
    from azure.cli.core import AzCommandsLoader
    from azure.cli.core.commands import AzCliCommand
    cli_ctx = DummyCli()
    loader = AzCommandsLoader(cli_ctx, resource_type=ResourceType.MGMT_APPSERVICE)
    cmd = AzCliCommand(loader, 'test', None)
    cmd.command_kwargs = {'resource_type': ResourceType.MGMT_APPSERVICE}
    cmd.cli_ctx = cli_ctx
    return cmd


class TestWebappMocked(unittest.TestCase):
    def setUp(self):
        self.client = WebSiteManagementClient(mock.MagicMock(), '123455678')

    @mock.patch('azure.cli.command_modules.appservice.custom._update_site_source_control_properties_for_gh_action')
    @mock.patch('azure.cli.command_modules.appservice.custom._add_publish_profile_to_github')
    @mock.patch('azure.cli.command_modules.appservice.custom.prompt_y_n')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_runtime_info')
    @mock.patch('github.Github')
    @mock.patch('azure.cli.command_modules.appservice.custom.parse_resource_id')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_site_availability')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_details')
    def test_webapp_github_actions_add(self, get_app_details_mock, web_client_factory_mock, site_availability_mock, *args):
        runtime = "python:3.9"
        rg = "group"
        is_linux = True
        cmd = _get_test_cmd()
        get_app_details_mock.return_value = mock.Mock()
        get_app_details_mock.return_value.resource_group = rg
        web_client_factory_mock.return_value.app_service_plans.get.return_value.reserved = is_linux
        site_availability_mock.return_value.name_available = False

        with mock.patch('azure.cli.command_modules.appservice.custom._runtime_supports_github_actions', autospec=True) as m:
            add_github_actions(cmd, rg, "name", "repo", runtime, "token")
            m.assert_called_with(cmd, runtime.replace(":", "|"), is_linux)

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    def test_set_deployment_user_creds(self, client_factory_mock):
        class MockClient:
            def update_publishing_user(self, user):
                # Don't do an actual call, just return the incoming user
                return user

        client_factory_mock.return_value = MockClient()

        # action
        user = set_deployment_user(_get_test_cmd(), 'admin', 'verySecret1')

        # assert things get wired up with a result returned
        assert user.publishing_user_name == 'admin'
        assert user.publishing_password == 'verySecret1'

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    def test_set_source_control_token(self, client_factory_mock):
        client = mock.Mock()
        client_factory_mock.return_value = client
        cmd_mock = _get_test_cmd()
        SourceControl, SourceControlProperties = cmd_mock.get_models('SourceControl', 'SourceControlProperties')
        sc = SourceControl(properties=SourceControlProperties(token='veryNiceToken'))
        client.update_source_control.return_value = sc

        # action
        result = update_git_token(cmd_mock, 'veryNiceToken')

        # assert things gets wired up
        self.assertEqual(result.token, None)

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    def test_set_domain_name(self, client_factory_mock):
        client = mock.Mock()
        client_factory_mock.return_value = client
        cmd_mock = _get_test_cmd()
        # set up the return value for getting a webapp
        Site, HostNameBinding = cmd_mock.get_models('Site', 'HostNameBinding')
        webapp = Site(location='westus')
        webapp.name = 'veryNiceWebApp'
        client.web_apps.get.return_value = webapp

        # set up the result value of putting a domain name
        domain = 'veryNiceDomain'
        binding = HostNameBinding(domain_id=domain,
                                  custom_host_name_dns_record_type='A',
                                  host_name_type='Managed')
        client.web_apps.create_or_update_host_name_binding.return_value = binding
        client.web_apps.create_or_update_host_name_binding_slot.return_value = binding
        # action
        result = add_hostname(cmd_mock, 'g1', webapp.name, domain)

        # assert
        self.assertEqual(result.domain_id, domain)

        # action- Slot
        result = add_hostname(cmd_mock, 'g1', webapp.name, domain, 'slot1')

        # assert
        self.assertEqual(result.domain_id, domain)

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    def test_get_external_ip_from_ase(self, client_factory_mock):
        client = mock.Mock()
        client_factory_mock.return_value = client
        cmd_mock = _get_test_cmd()
        # set up the web inside a ASE, with an ip based ssl binding
        HostingEnvironmentProfile = cmd_mock.get_models('HostingEnvironmentProfile')
        host_env = HostingEnvironmentProfile(id='id11')
        host_env.name = 'ase1'
        host_env.resource_group = 'myRg'

        HostNameSslState, SslState, Site, AddressResponse = \
            cmd_mock.get_models('HostNameSslState', 'SslState', 'Site', 'AddressResponse')

        host_ssl_state = HostNameSslState(ssl_state=SslState.ip_based_enabled, virtual_ip='1.2.3.4')
        client.web_apps.get.return_value = Site(name='antarctica', hosting_environment_profile=host_env,
                                                host_name_ssl_states=[host_ssl_state], location='westus')
        client.app_service_environments.list_vips.return_value = AddressResponse()

        # action
        result = get_external_ip(cmd_mock, 'myRg', 'myWeb')

        # assert, we return the virtual ip from the ip based ssl binding
        self.assertEqual('1.2.3.4', result['ip'])

        # tweak to have no ip based ssl binding, but it is in an internal load balancer
        host_ssl_state2 = HostNameSslState(ssl_state=SslState.sni_enabled)
        client.web_apps.get.return_value = Site(name='antarctica', hosting_environment_profile=host_env,
                                                host_name_ssl_states=[host_ssl_state2], location='westus')
        client.app_service_environments.list_vips.return_value = AddressResponse(internal_ip_address='4.3.2.1')

        # action
        result = get_external_ip(cmd_mock, 'myRg', 'myWeb')

        # assert, we take the ILB address
        self.assertEqual('4.3.2.1', result['ip'])

        # tweak to have no ip based ssl binding, and not in internal load balancer
        host_ssl_state2 = HostNameSslState(ssl_state=SslState.sni_enabled)
        client.web_apps.get.return_value = Site(name='antarctica', hosting_environment_profile=host_env,
                                                host_name_ssl_states=[host_ssl_state2], location='westus')
        client.app_service_environments.list_vips.return_value = AddressResponse(service_ip_address='1.1.1.1')

        # action
        result = get_external_ip(cmd_mock, 'myRg', 'myWeb')

        # assert, we take service ip
        self.assertEqual('1.1.1.1', result['ip'])

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._resolve_hostname_through_dns', autospec=True)
    def test_get_external_ip_from_dns(self, resolve_hostname_mock, client_factory_mock):
        client = mock.Mock()
        client_factory_mock.return_value = client
        cmd_mock = _get_test_cmd()
        # set up the web inside a ASE, with an ip based ssl binding
        Site = cmd_mock.get_models('Site')
        site = Site(name='antarctica', location='westus')
        site.default_host_name = 'myweb.com'
        client.web_apps.get.return_value = site

        # action
        get_external_ip(mock.MagicMock(), 'myRg', 'myWeb')

        # assert, we return the virtual ip from the ip based ssl binding
        resolve_hostname_mock.assert_called_with('myweb.com')

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.is_centauri_functionapp', autospec=True)
    def test_update_site_config(self, is_centauri_functionapp_mock, site_op_mock):

        cmd_mock = _get_test_cmd()
        SiteConfig = cmd_mock.get_models('SiteConfig')
        site_config = SiteConfig()
        site_op_mock.return_value = site_config

        is_centauri_functionapp_mock.return_value = False
        # action
        update_site_configs(cmd_mock, 'myRG', 'myweb', java_version='1.8')
        # assert
        self.assertEqual(site_config.java_version, '1.8')
        # point check some unrelated properties should stay at None
        self.assertEqual(site_config.use32_bit_worker_process, None)
        self.assertEqual(site_config.java_container, None)

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    def test_list_publish_profiles_on_slots(self, site_op_mock):
        site_op_mock.return_value = [b'<publishData><publishProfile publishUrl="ftp://123"/><publishProfile publishUrl="ftp://1234"/></publishData>']
        # action
        result = list_publish_profiles(mock.MagicMock(), 'myRG', 'myweb', 'slot1')
        # assert
        site_op_mock.assert_called_with(mock.ANY, 'myRG', 'myweb', 'list_publishing_profile_xml_with_secrets', 'slot1',
                                        {'format': 'WebDeploy'})
        self.assertTrue(result[0]['publishUrl'].startswith('ftp://123'))

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.get_streaming_log', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.open_page_in_browser', autospec=True)
    def test_browse_with_trace(self, webbrowser_mock, log_mock, site_op_mock):
        cmd_mock = _get_test_cmd()
        Site, HostNameSslState, SslState = cmd_mock.get_models('Site', 'HostNameSslState', 'SslState')
        site = Site(location='westus', name='antarctica')
        site.default_host_name = 'haha.com'
        site.enabled_host_names = [site.default_host_name]
        site.host_name_ssl_states = [HostNameSslState(name='does not matter',
                                                      ssl_state=SslState.ip_based_enabled)]

        site_op_mock.return_value = site
        # action
        view_in_browser(mock.MagicMock(), 'myRG', 'myweb', logs=True)
        # assert
        webbrowser_mock.assert_called_with('https://haha.com')
        log_mock.assert_called_with(mock.ANY, 'myRG', 'myweb', None, None)

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.is_centauri_functionapp', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._rename_server_farm_props', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._fill_ftp_publishing_url', autospec=True)
    def test_show_webapp(self, file_ftp_mock, rename_mock, is_centauri_functionapp_mock, site_op_mock):
        faked_web = mock.MagicMock()
        site_op_mock.return_value = faked_web
        is_centauri_functionapp_mock.return_value = False
        # action
        result = show_app(mock.MagicMock(), 'myRG', 'myweb', slot=None)
        # assert (we invoke the site op)
        self.assertEqual(faked_web, result)
        self.assertTrue(rename_mock.called)
        self.assertTrue(file_ftp_mock.called)

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    def test_sync_repository_skip_bad_error(self, site_op_mock):
        resp = FakedResponse(200)  # because of bad spec, sdk throws on 200.
        setattr(resp, 'reason', 'bad error')
        site_op_mock.side_effect = HttpResponseError(response=resp)
        # action
        sync_site_repo(mock.MagicMock(), 'myRG', 'myweb')
        # assert
        pass  # if we are here, it means CLI has captured the bogus exception

    def test_match_host_names_from_cert(self):
        result = _match_host_names_from_cert(['*.mysite.com'], ['admin.mysite.com', 'log.mysite.com', 'mysite.com'])
        self.assertEqual(set(['admin.mysite.com', 'log.mysite.com']), result)

        result = _match_host_names_from_cert(['*.mysite.com', 'mysite.com'], ['admin.mysite.com', 'log.mysite.com', 'mysite.com'])
        self.assertEqual(set(['admin.mysite.com', 'log.mysite.com', 'mysite.com']), result)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers', return_value={"auth": "1245!"})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.ping_site', autospec=True)
    @mock.patch('threading.Thread', autospec=True)
    def test_log_stream_supply_cli_ctx(self, threading_mock, ping_site_mock, get_scm_url_mock, get_scm_site_headers_mock):

        # test exception to exit the streaming loop
        class ErrorToExitInfiniteLoop(Exception):
            pass

        threading_mock.side_effect = ErrorToExitInfiniteLoop('Expected error to exit early')
        get_scm_url_mock.return_value = 'http://great_url'
        ping_site_mock.return_value = None
        cmd_mock = mock.MagicMock()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock
        rg_name = "rg"
        app_name = "web1"

        try:
            # action
            get_streaming_log(cmd_mock, rg_name, app_name)
            self.fail('test exception was not thrown')
        except ErrorToExitInfiniteLoop:
            # assert
            get_scm_site_headers_mock.assert_called_with(cli_ctx_mock, app_name, rg_name, None)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_url', autospec=True)
    def test_log_stream_ping_site_failed(self, get_site_url_mock):
        import urllib3
        get_site_url_mock.return_value = 'http://unreachable-url'
        cmd_mock = mock.MagicMock()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock
        rg_name = "rg"
        app_name = "web1"

        try:
            # action
            get_streaming_log(cmd_mock, rg_name, app_name)
            self.fail('Exception not thrown even when site ping should fail')
        except urllib3.exceptions.MaxRetryError:
            # assert
            get_site_url_mock.assert_called_with(cmd_mock, rg_name, app_name, None)

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    def test_restore_deleted_webapp(self, site_op_mock):
        cmd_mock = mock.MagicMock()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock
        DeletedAppRestoreRequest = cmd_mock.get_models('DeletedAppRestoreRequest')
        request = DeletedAppRestoreRequest(deleted_site_id='12345', recover_configuration=False)

        # action
        restore_deleted_webapp(cmd_mock, '12345', 'rg', 'web1', None, True)

        # assert
        site_op_mock.assert_called_with(cli_ctx_mock, 'rg', 'web1', 'begin_restore_from_deleted_app', None, request)

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    def test_list_webapp_snapshots(self, site_op_mock):
        cmd_mock = mock.MagicMock()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        # action
        list_snapshots(cmd_mock, 'rg', 'web1', None)

        # assert
        site_op_mock.assert_called_with(cli_ctx_mock, 'rg', 'web1', 'list_snapshots', None)

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    def test_restore_snapshot(self, generic_site_op_mock, client_factory_mock):
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cli_ctx_mock.data = {'subscription_id': 'sub1'}
        cmd_mock.cli_ctx = cli_ctx_mock

        client = mock.MagicMock()
        client.web_apps.restore_snapshot_slot = mock.MagicMock()
        client.web_apps.restore_snapshot = mock.MagicMock()

        Site = cmd_mock.get_models('Site')
        site = Site(name='src_web', location='location')
        site.slot_name = 'src_slot'
        site.resouce_group = 'src_rg'
        site.id = '/subscriptions/sub1/resourceGroups/src_rg/providers/Microsoft.Web/sites/src_web/slots/src_slot'

        generic_site_op_mock.return_value = site

        client_factory_mock.return_value = client
        


        SnapshotRecoverySource, SnapshotRestoreRequest = \
            cmd_mock.get_models('SnapshotRecoverySource', 'SnapshotRestoreRequest')
        source = SnapshotRecoverySource(id='/subscriptions/sub1/resourceGroups/src_rg/providers/Microsoft.Web/sites/src_web/slots/src_slot', location='location')
        request = SnapshotRestoreRequest(overwrite=False, snapshot_time='2018-12-07T02:01:31.4708832Z',
                                         recovery_source=source, recover_configuration=False)
        overwrite_request = SnapshotRestoreRequest(overwrite=True, snapshot_time='2018-12-07T02:01:31.4708832Z', recover_configuration=True)

        # action
        restore_snapshot(cmd_mock, 'rg', 'web1', '2018-12-07T02:01:31.4708832Z', slot='slot1', restore_content_only=True,
                         source_resource_group='src_rg', source_name='src_web', source_slot='src_slot')
        restore_snapshot(cmd_mock, 'rg', 'web1', '2018-12-07T02:01:31.4708832Z', restore_content_only=False)

        # assert
        client.web_apps.begin_restore_snapshot_slot.assert_called_with('rg', 'web1', 'slot1', request)
        client.web_apps.begin_restore_snapshot.assert_called_with('rg', 'web1', overwrite_request)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers', return_value={"auth": "1245!"})
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._get_log', autospec=True)
    def test_download_log_supply_cli_ctx(self, get_log_mock, get_scm_url_mock, site_op_mock, *args):
        def test_result():
            res = mock.MagicMock()
            res.publishing_user_name, res.publishing_password = 'great_user', 'secret_password'
            return res
        test_scm_url = 'http://great_url'
        get_scm_url_mock.return_value = test_scm_url
        publish_cred_mock = mock.MagicMock()
        publish_cred_mock.result = test_result
        site_op_mock.return_value = publish_cred_mock
        cmd_mock = mock.MagicMock()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        # action
        download_historical_logs(cmd_mock, 'rg', 'web1')

        # assert
        get_log_mock.assert_called_with(test_scm_url + '/dump', {"auth": "1245!"}, None)

    def test_valid_linux_create_options(self):
        some_runtime = 'TOMCAT|8.5-jre8'
        test_docker_image = 'lukasz/great-image:123'
        test_multi_container_config = 'some_config.yaml'
        test_multi_container_type = 'COMPOSE'

        self.assertTrue(validate_container_app_create_options(some_runtime, None, None, None))
        self.assertTrue(validate_container_app_create_options(None, test_docker_image, None, None))
        self.assertTrue(validate_container_app_create_options(None, None, test_multi_container_config, test_multi_container_type))
        self.assertFalse(validate_container_app_create_options(some_runtime, None, test_multi_container_config, test_multi_container_type))
        self.assertFalse(validate_container_app_create_options(some_runtime, None, test_multi_container_config, None))
        self.assertFalse(validate_container_app_create_options(some_runtime, test_docker_image, test_multi_container_config, None))
        self.assertFalse(validate_container_app_create_options(None, None, test_multi_container_config, None))
        self.assertFalse(validate_container_app_create_options(None, None, None, None))

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._StackRuntimeHelper', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.get_site_availability', autospec=True)
    def test_linux_webapp_create_no_runtime_raises_error(self, get_site_avail_mock,
                                                         stack_helper_mock, web_client_mock):
        cmd_mock = _get_test_cmd()
        SiteConfig, SkuDescription, NameValuePair = cmd_mock.get_models(
            'SiteConfig', 'SkuDescription', 'NameValuePair')
        cmd_mock.get_models = mock.MagicMock(return_value=(SiteConfig, SkuDescription, NameValuePair))

        # Mock a Linux plan (reserved=True)
        plan_info = mock.MagicMock()
        plan_info.reserved = True
        plan_info.location = 'eastus'
        plan_info.id = '/subscriptions/sub/resourceGroups/rg/providers/Microsoft.Web/serverfarms/plan'
        plan_info.sku = SkuDescription(name='F1')
        web_client_mock.return_value.app_service_plans.get.return_value = plan_info

        # Mock site availability (new app name)
        name_validation = mock.MagicMock()
        name_validation.name_available = True
        get_site_avail_mock.return_value = name_validation

        with self.assertRaises(ArgumentUsageError) as context:
            create_webapp(cmd_mock, 'test-rg', 'test-app', 'test-plan')

        self.assertIn('Creating a Linux webapp requires one of the following', str(context.exception))
        self.assertIn('--runtime', str(context.exception))
        self.assertIn('--os-type linux', str(context.exception))

    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._verify_hostname_binding', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    def test_create_managed_ssl_cert(self, generic_site_op_mock, client_factory_mock, verify_binding_mock, is_flex_mock):
        webapp_name = 'someWebAppName'
        rg_name = 'someRgName'
        farm_id = '/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/rg1/providers/Microsoft.Web/serverfarms/farm1'
        host_name = 'www.contoso.com'

        client = mock.MagicMock()
        client_factory_mock.return_value = client
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cli_ctx_mock.data = {'subscription_id': 'sub1'}
        cmd_mock.cli_ctx = cli_ctx_mock
        Site, Certificate = cmd_mock.get_models('Site', 'Certificate')
        site = Site(name=webapp_name, location='westeurope')
        site.server_farm_id = farm_id
        generic_site_op_mock.return_value = site
        is_flex_mock.return_value = False

        verify_binding_mock.return_value = False
        with self.assertRaises(CLIError):
            create_managed_ssl_cert(cmd_mock, rg_name, webapp_name, host_name, None)

        verify_binding_mock.return_value = True
        create_managed_ssl_cert(cmd_mock, rg_name, webapp_name, host_name, None)

        cert_def = Certificate(location='westeurope', canonical_name=host_name,
                               server_farm_id=farm_id, password='')
        client.certificates.create_or_update.assert_called_once_with(name=host_name, resource_group_name=rg_name,
                                                                     certificate_envelope=cert_def)


    def test_update_app_settings_error_handling_no_parameters(self):
        """Test that MutuallyExclusiveArgumentError is raised when neither settings nor slot_settings are provided."""
        cmd_mock = _get_test_cmd()
        
        # Test missing both parameters - should fail early without calling any services
        with self.assertRaisesRegex(MutuallyExclusiveArgumentError, 
                                   "Please provide either --settings or --slot-settings parameter"):
            update_app_settings(cmd_mock, 'test-rg', 'test-app')

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom.shell_safe_json_parse')
    def test_update_app_settings_error_handling_invalid_format(self, mock_json_parse, mock_site_op):
        """Test that InvalidArgumentValueError is raised for invalid setting formats."""
        cmd_mock = _get_test_cmd()
        
        # Setup minimal mocks needed to reach the error handling code
        mock_app_settings = mock.MagicMock()
        mock_app_settings.properties = {}
        mock_site_op.return_value = mock_app_settings
        
        # Mock shell_safe_json_parse to raise InvalidArgumentValueError (simulating invalid JSON)
        mock_json_parse.side_effect = InvalidArgumentValueError("Invalid JSON format")
        
        # Test invalid format that can't be parsed as JSON or key=value
        invalid_setting = "invalid_format_no_equals_no_json"
        expected_message = r"Invalid setting format.*Expected 'key=value' format or valid JSON"
        
        with self.assertRaisesRegex(InvalidArgumentValueError, expected_message):
            update_app_settings(cmd_mock, 'test-rg', 'test-app', settings=[invalid_setting])

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom.shell_safe_json_parse')
    def test_update_app_settings_error_handling_invalid_format_no_equals(self, mock_json_parse, mock_site_op):
        """Test ValueError path when shell_safe_json_parse raises InvalidArgumentValueError and string contains no '='."""
        cmd_mock = _get_test_cmd()
        
        # Setup minimal mocks needed to reach the error handling code
        mock_app_settings = mock.MagicMock()
        mock_app_settings.properties = {}
        mock_site_op.return_value = mock_app_settings
        
        # Mock shell_safe_json_parse to raise InvalidArgumentValueError
        mock_json_parse.side_effect = InvalidArgumentValueError("Invalid JSON format")
        
        # Test invalid format with no equals sign - this should trigger ValueError in split('=', 1)
        invalid_setting_no_equals = "invalidformatthatcontainsnoequalsign"
        expected_message = r"Invalid setting format.*Expected 'key=value' format or valid JSON"
        
        with self.assertRaisesRegex(InvalidArgumentValueError, expected_message):
            update_app_settings(cmd_mock, 'test-rg', 'test-app', settings=[invalid_setting_no_equals])

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_centauri_functionapp')
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_settings_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom._build_app_settings_output')
    def test_update_app_settings_success_key_value_format(self, mock_build, mock_settings_op, mock_centauri, 
                                                         mock_client_factory, mock_site_op):
        """Test successful processing of key=value format settings."""
        cmd_mock = _get_test_cmd()
        
        # Setup mocks
        mock_app_settings = mock.MagicMock()
        mock_app_settings.properties = {}
        mock_site_op.return_value = mock_app_settings
        
        mock_client = mock.MagicMock()
        mock_client_factory.return_value = mock_client
        mock_centauri.return_value = False
        mock_settings_op.return_value = mock_app_settings
        mock_build.return_value = {"KEY1": "value1", "KEY2": "value2"}
        
        # Test valid key=value format
        result = update_app_settings(cmd_mock, 'test-rg', 'test-app', 
                                   settings=['KEY1=value1', 'KEY2=value2'])
        
        # Verify the function completed successfully
        self.assertEqual(result["KEY1"], "value1")
        self.assertEqual(result["KEY2"], "value2")
        mock_build.assert_called_once()

    @mock.patch('azure.cli.command_modules.appservice.custom.send_raw_request')
    def test_update_application_settings_polling_error_handling(self, mock_send_request):
        """Test that AzureResponseError is raised in polling function when appropriate."""
        cmd_mock = _get_test_cmd()
        
        # Mock an exception that doesn't have the expected structure
        class MockException(Exception):
            def __init__(self):
                self.response = mock.MagicMock()
                self.response.status_code = 400  # Not 202
                self.response.headers = {}
        
        # Mock _generic_settings_operation to raise the exception
        with mock.patch('azure.cli.command_modules.appservice.custom._generic_settings_operation') as mock_settings_op, \
             self.assertRaisesRegex(AzureResponseError, "Failed to update application settings"):
            mock_settings_op.side_effect = MockException()
            update_application_settings_polling(cmd_mock, 'test-rg', 'test-app', 
                                               mock.MagicMock(), None, mock.MagicMock())

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_centauri_functionapp')
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_settings_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom._build_app_settings_output')
    def test_update_app_settings_success_with_slot_settings(self, mock_build, mock_settings_op, mock_centauri,
                                                           mock_client_factory, mock_site_op):
        """Test successful processing with slot settings."""
        cmd_mock = _get_test_cmd()
        
        # Setup mocks
        mock_app_settings = mock.MagicMock()
        mock_app_settings.properties = {}
        mock_site_op.return_value = mock_app_settings
        
        mock_client = mock.MagicMock()
        mock_slot_config = mock.MagicMock()
        mock_slot_config.app_setting_names = []
        mock_client.web_apps.list_slot_configuration_names.return_value = mock_slot_config
        mock_client_factory.return_value = mock_client
        mock_centauri.return_value = False
        mock_settings_op.return_value = mock_app_settings
        mock_build.return_value = {"SLOT_KEY": "slot_value"}
        
        # Test with slot settings
        result = update_app_settings(cmd_mock, 'test-rg', 'test-app', 
                                   settings=['REGULAR_KEY=regular_value'],
                                   slot_settings=['SLOT_KEY=slot_value'])
        
        # Verify slot configuration was updated
        mock_client.web_apps.list_slot_configuration_names.assert_called_once()
        mock_client.web_apps.update_slot_configuration_names.assert_called_once()
        mock_build.assert_called_once()


class TestUpdateWebapp(unittest.TestCase):

    def _create_site_instance(self, cmd):
        Site = cmd.get_models('Site')
        SiteConfig = cmd.get_models('SiteConfig')
        site_config = SiteConfig(number_of_workers=1)
        instance = Site(location='eastus', site_config=site_config)
        instance.kind = 'app,linux'
        return instance

    def test_update_webapp_platform_release_channel_extended(self):
        cmd_mock = _get_test_cmd()
        instance = self._create_site_instance(cmd_mock)

        result = update_webapp(cmd_mock, instance, platform_release_channel='Extended')

        self.assertEqual(result.properties["platformReleaseChannel"], "Extended")

    def test_update_webapp_platform_release_channel_standard(self):
        cmd_mock = _get_test_cmd()
        instance = self._create_site_instance(cmd_mock)

        result = update_webapp(cmd_mock, instance, platform_release_channel='Standard')

        self.assertEqual(result.properties["platformReleaseChannel"], "Standard")

    def test_update_webapp_platform_release_channel_latest(self):
        cmd_mock = _get_test_cmd()
        instance = self._create_site_instance(cmd_mock)

        result = update_webapp(cmd_mock, instance, platform_release_channel='Latest')

        self.assertEqual(result.properties["platformReleaseChannel"], "Latest")


class TestStartupLogsMocked(unittest.TestCase):
    """Tests for az webapp log startup list/show commands."""

    def setUp(self):
        # Default: pretend the app is Linux so existing tests exercise the happy path.
        # Individual tests can re-patch these when they need different behavior.
        is_linux_patch = mock.patch(
            'azure.cli.command_modules.appservice.custom.is_linux_webapp',
            return_value=True)
        client_factory_patch = mock.patch(
            'azure.cli.command_modules.appservice.custom.web_client_factory')
        is_linux_patch.start()
        client_factory_patch.start()
        self.addCleanup(is_linux_patch.stop)
        self.addCleanup(client_factory_patch.stop)

    def _make_response(self, status_code=200, json_data=None, text='', headers=None, reason=''):
        resp = mock.MagicMock()
        resp.status_code = status_code
        resp.reason = reason
        resp.text = text
        resp.headers = headers or {}
        resp.json.return_value = json_data
        return resp

    # ---- list_startup_logs ----

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_list_startup_logs_success(self, requests_get_mock, _scm_url_mock, _headers_mock):
        files = [
            {'Filename': '2026_04_13_lw0sdlwk000002_success.log', 'Href': '/api/startuplogs/...'},
            {'Filename': '2026_04_13_lw0sdlwk000003_failure.log', 'Href': '/api/startuplogs/...'},
        ]
        requests_get_mock.return_value = self._make_response(200, json_data={'files': files})

        result = list_startup_logs(_get_test_cmd(), 'myRG', 'myApp')

        self.assertEqual(result, files)
        requests_get_mock.assert_called_once_with(
            'https://myapp.scm.azurewebsites.net/api/startuplogs',
            headers={'Authorization': 'Bearer token'},
            params={}
        )

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_list_startup_logs_with_filters(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(200, json_data={'files': []})

        list_startup_logs(_get_test_cmd(), 'myRG', 'myApp', outcome='failure', instance='lw0sdlwk000002')

        requests_get_mock.assert_called_once_with(
            'https://myapp.scm.azurewebsites.net/api/startuplogs',
            headers={'Authorization': 'Bearer token'},
            params={'type': 'failure', 'instance': 'lw0sdlwk000002'}
        )

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_list_startup_logs_404_graceful(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(404)

        with mock.patch('azure.cli.command_modules.appservice.custom.logger') as logger_mock:
            result = list_startup_logs(_get_test_cmd(), 'myRG', 'myApp')

        self.assertEqual(result, [])
        logger_mock.warning.assert_called_once()
        self.assertIn('platform version', logger_mock.warning.call_args[0][0])

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_list_startup_logs_500_raises(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(500, reason='Internal Server Error')

        with self.assertRaises(CLIError) as cm:
            list_startup_logs(_get_test_cmd(), 'myRG', 'myApp')
        self.assertIn('500', str(cm.exception))

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_list_startup_logs_with_slot(self, requests_get_mock, scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(200, json_data={'files': []})

        list_startup_logs(_get_test_cmd(), 'myRG', 'myApp', slot='staging')

        scm_url_mock.assert_called_once_with(mock.ANY, 'myRG', 'myApp', 'staging')

    # ---- show_startup_log ----

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_latest(self, requests_get_mock, _scm_url_mock, _headers_mock):
        log_text = 'Container started successfully.\nListening on port 8080.'
        requests_get_mock.return_value = self._make_response(
            200, text=log_text,
            headers={
                'Content-Type': 'text/plain',
                'X-StartupLog-Filename': '2026_04_13_lw0_success.log',
                'X-StartupLog-Date': '2026-04-13T10:00:00Z',
                'X-StartupLog-Instance': 'lw0sdlwk000002',
                'X-StartupLog-Outcome': 'success',
            }
        )

        result = show_startup_log(_get_test_cmd(), 'myRG', 'myApp')

        self.assertEqual(result['content'], log_text)
        self.assertEqual(result['filename'], '2026_04_13_lw0_success.log')
        self.assertEqual(result['instance'], 'lw0sdlwk000002')
        self.assertEqual(result['outcome'], 'success')
        requests_get_mock.assert_called_once_with(
            'https://myapp.scm.azurewebsites.net/api/startuplogs?latest=true',
            headers={'Authorization': 'Bearer token'}
        )

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_specific_filename(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(
            200, text='log content',
            headers={'Content-Type': 'text/plain'}
        )

        show_startup_log(_get_test_cmd(), 'myRG', 'myApp', filename='2026_04_13_lw0_success.log')

        requests_get_mock.assert_called_once_with(
            'https://myapp.scm.azurewebsites.net/api/startuplogs/2026_04_13_lw0_success.log',
            headers={'Authorization': 'Bearer token'}
        )

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_with_instance(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(
            200, text='instance log content',
            headers={
                'Content-Type': 'text/plain',
                'X-StartupLog-Filename': '2026_04_13_lw0sdlwk000002_failure.log',
                'X-StartupLog-Instance': 'lw0sdlwk000002',
                'X-StartupLog-Outcome': 'failure',
            }
        )

        result = show_startup_log(_get_test_cmd(), 'myRG', 'myApp', instance='lw0sdlwk000002')

        self.assertEqual(result['content'], 'instance log content')
        self.assertEqual(result['instance'], 'lw0sdlwk000002')
        requests_get_mock.assert_called_once_with(
            'https://myapp.scm.azurewebsites.net/api/startuplogs?latest=true&instance=lw0sdlwk000002',
            headers={'Authorization': 'Bearer token'}
        )

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_404_no_filename(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(404)

        with mock.patch('azure.cli.command_modules.appservice.custom.logger') as logger_mock:
            result = show_startup_log(_get_test_cmd(), 'myRG', 'myApp')

        self.assertIsNone(result)
        self.assertIn('platform version', logger_mock.warning.call_args[0][0])

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_404_with_filename(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(404)

        with mock.patch('azure.cli.command_modules.appservice.custom.logger') as logger_mock:
            result = show_startup_log(_get_test_cmd(), 'myRG', 'myApp', filename='nonexistent.log')

        self.assertIsNone(result)
        self.assertIn('nonexistent.log', logger_mock.warning.call_args[0][1])

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_500_raises(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(500, reason='Internal Server Error')

        with self.assertRaises(CLIError) as cm:
            show_startup_log(_get_test_cmd(), 'myRG', 'myApp')
        self.assertIn('500', str(cm.exception))

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_json_response(self, requests_get_mock, _scm_url_mock, _headers_mock):
        json_data = {'filename': 'test.log', 'content': 'data'}
        requests_get_mock.return_value = self._make_response(
            200, json_data=json_data,
            headers={'Content-Type': 'application/json'}
        )

        result = show_startup_log(_get_test_cmd(), 'myRG', 'myApp')

        self.assertEqual(result, json_data)

    # ---- Linux-only gating ----

    def test_list_startup_logs_raises_on_windows(self):
        with mock.patch('azure.cli.command_modules.appservice.custom.is_linux_webapp',
                        return_value=False):
            with self.assertRaises(ArgumentUsageError) as cm:
                list_startup_logs(_get_test_cmd(), 'myRG', 'myWindowsApp')
        self.assertIn('Linux', str(cm.exception))

    def test_show_startup_log_raises_on_windows(self):
        with mock.patch('azure.cli.command_modules.appservice.custom.is_linux_webapp',
                        return_value=False):
            with self.assertRaises(ArgumentUsageError) as cm:
                show_startup_log(_get_test_cmd(), 'myRG', 'myWindowsApp')
        self.assertIn('Linux', str(cm.exception))

    # ---- --filename / --instance mutual exclusion ----

    def test_show_startup_log_filename_and_instance_mutually_exclusive(self):
        with self.assertRaises(MutuallyExclusiveArgumentError) as cm:
            show_startup_log(_get_test_cmd(), 'myRG', 'myApp',
                             filename='2026_04_13_lw0_success.log',
                             instance='lw0sdlwk000002')
        self.assertIn('--filename', str(cm.exception))
        self.assertIn('--instance', str(cm.exception))

    # ---- 404 disambiguation when --instance is set ----

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_list_startup_logs_404_with_instance(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(404)

        with mock.patch('azure.cli.command_modules.appservice.custom.logger') as logger_mock:
            result = list_startup_logs(_get_test_cmd(), 'myRG', 'myApp', instance='lw0sdlwk000002')

        self.assertEqual(result, [])
        logger_mock.warning.assert_called_once()
        self.assertIn('instance', logger_mock.warning.call_args[0][0])
        self.assertEqual(logger_mock.warning.call_args[0][1], 'lw0sdlwk000002')

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer token'})
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('requests.get')
    def test_show_startup_log_404_with_instance(self, requests_get_mock, _scm_url_mock, _headers_mock):
        requests_get_mock.return_value = self._make_response(404)

        with mock.patch('azure.cli.command_modules.appservice.custom.logger') as logger_mock:
            result = show_startup_log(_get_test_cmd(), 'myRG', 'myApp', instance='lw0sdlwk000002')

        self.assertIsNone(result)
        logger_mock.warning.assert_called_once()
        self.assertIn('instance', logger_mock.warning.call_args[0][0])
        self.assertEqual(logger_mock.warning.call_args[0][1], 'lw0sdlwk000002')


class TestRuntimeFailedHintMocked(unittest.TestCase):
    """Tests that the TIP hint appears in RuntimeFailed and timeout errors."""

    def _make_deployment_response(self, status, num_in_progress=0, num_successful=0,
                                  num_failed=1, errors=None, failure_logs=None):
        return {
            'properties': {
                'status': status,
                'numberOfInstancesInProgress': str(num_in_progress),
                'numberOfInstancesSuccessful': str(num_successful),
                'numberOfInstancesFailed': str(num_failed),
                'errors': errors or [],
                'failedInstancesLogs': failure_logs,
            }
        }

    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('azure.cli.command_modules.appservice.custom.time.sleep')
    @mock.patch('azure.cli.command_modules.appservice.custom.time.time')
    @mock.patch('azure.cli.command_modules.appservice.custom.send_raw_request')
    def test_runtime_failed_includes_startup_log_hint(self, send_raw_mock, time_mock,
                                                      sleep_mock, _scm_url_mock):
        from azure.cli.command_modules.appservice.custom import _poll_deployment_runtime_status

        time_mock.return_value = 10  # constant — never times out, RuntimeFailed triggers on first iteration
        resp_mock = mock.MagicMock()
        resp_mock.json.return_value = self._make_deployment_response('RuntimeFailed')
        send_raw_mock.return_value = resp_mock

        with self.assertRaises(CLIError) as cm:
            _poll_deployment_runtime_status(
                _get_test_cmd(), 'myRG', 'myApp', None,
                'https://management.azure.com/deploymentstatus', 'deploy-id-1'
            )

        error_msg = str(cm.exception)
        self.assertIn('az webapp log startup show -n myApp -g myRG', error_msg)
        self.assertIn('failed to start', error_msg)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    @mock.patch('azure.cli.command_modules.appservice.custom.time.sleep')
    @mock.patch('azure.cli.command_modules.appservice.custom.time.time')
    @mock.patch('azure.cli.command_modules.appservice.custom.send_raw_request')
    def test_timeout_includes_startup_log_hint(self, send_raw_mock, time_mock,
                                               sleep_mock, _scm_url_mock):
        from azure.cli.command_modules.appservice.custom import _poll_deployment_runtime_status
        import itertools

        # Advancing counter: each call returns 0, 1, 2, ... — exceeds timeout=1 after first loop
        time_mock.side_effect = itertools.count(0)
        resp_mock = mock.MagicMock()
        resp_mock.json.return_value = self._make_deployment_response('RuntimeStarting')
        send_raw_mock.return_value = resp_mock

        with self.assertRaises(CLIError) as cm:
            _poll_deployment_runtime_status(
                _get_test_cmd(), 'myRG', 'myApp', None,
                'https://management.azure.com/deploymentstatus', 'deploy-id-1',
                timeout=1
            )

        error_msg = str(cm.exception)
        self.assertIn('az webapp log startup show -n myApp -g myRG', error_msg)
        self.assertIn('Timeout', error_msg)


class FakedResponse:  # pylint: disable=too-few-public-methods
    def __init__(self, status_code):
        self.status_code = status_code


class TestCreateAppServicePlanDefaults(unittest.TestCase):
    """Tests for create_app_service_plan default SKU behavior"""

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_location_from_resource_group', return_value='eastus')
    def test_default_sku_is_p0v3_when_not_specified(self, mock_location, mock_client_factory):
        from azure.cli.command_modules.appservice.custom import create_app_service_plan
        mock_cmd = mock.MagicMock()
        mock_cmd.get_models.return_value = (mock.MagicMock(), mock.MagicMock(), mock.MagicMock())
        mock_cmd.cli_ctx = mock.MagicMock()
        mock_client = mock.MagicMock()
        mock_client_factory.return_value = mock_client

        # Call without sku parameter — should default to P0V3
        try:
            create_app_service_plan(mock_cmd, 'rg', 'plan', is_linux=True, hyper_v=False)
        except Exception:
            pass  # We don't care about downstream errors, just checking the SKU

        # Verify SkuDescription was called with P0V3 tier/name
        sku_description_cls = mock_cmd.get_models.return_value[1]
        sku_description_cls.assert_called()
        call_kwargs = sku_description_cls.call_args
        # The sku name should be normalized P0V3
        self.assertIn('P0V3', str(call_kwargs))


class TestOneDeployScmCache(unittest.TestCase):
    """Tests for the per-invocation SCM URL / SCM headers cache on OneDeployParams.

    The cache avoids duplicate `_get_scm_url` and `get_scm_site_headers` round
    trips between the publish leg and the status-poll leg of a single
    `az webapp deploy` invocation. See _perform_onedeploy_internal +
    _check_zip_deployment_status in custom.py.
    """

    def _make_params(self):
        from azure.cli.command_modules.appservice.custom import OneDeployParams
        params = OneDeployParams()
        params.cmd = mock.MagicMock()
        params.cmd.cli_ctx = mock.MagicMock()
        params.cmd.cli_ctx.data = {'headers': {'x-ms-client-request-id': 'req-1'}}
        params.resource_group_name = 'myRG'
        params.webapp_name = 'myApp'
        params.slot = None
        return params

    def test_get_or_fetch_scm_url_derives_from_cached_site(self):
        from azure.cli.command_modules.appservice.custom import _get_or_fetch_scm_url
        from azure.mgmt.web.models import HostType
        params = self._make_params()
        # Simulate a cached Site with a repository host
        repo_host = mock.MagicMock()
        repo_host.host_type = HostType.repository
        repo_host.name = 'myapp.scm.azurewebsites.net'
        std_host = mock.MagicMock()
        std_host.host_type = HostType.standard
        std_host.name = 'myapp.azurewebsites.net'
        params._cached_site = mock.MagicMock()
        params._cached_site.host_name_ssl_states = [std_host, repo_host]

        first = _get_or_fetch_scm_url(params)
        second = _get_or_fetch_scm_url(params)

        self.assertEqual(first, 'https://myapp.scm.azurewebsites.net')
        self.assertEqual(second, first)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url',
                return_value='https://myapp.scm.azurewebsites.net')
    def test_get_or_fetch_scm_url_falls_back_when_no_cached_site(self, get_scm_url_mock):
        from azure.cli.command_modules.appservice.custom import _get_or_fetch_scm_url
        params = self._make_params()
        params._cached_site = None

        result = _get_or_fetch_scm_url(params)

        self.assertEqual(result, 'https://myapp.scm.azurewebsites.net')
        get_scm_url_mock.assert_called_once_with(params.cmd, 'myRG', 'myApp', None)

    def test_populate_cached_scm_headers_basic_auth_lowercase_key(self):
        # The basic-auth branch of get_scm_site_headers builds headers via
        # urllib3.util.make_headers(basic_auth=...), which uses a lowercase
        # 'authorization' key (verified with urllib3 in CI). The cache must
        # match case-insensitively so the customer's actual Windows + basic
        # auth code path is covered.
        from azure.cli.command_modules.appservice.custom import _populate_cached_scm_headers
        params = self._make_params()
        headers = {
            'authorization': 'Basic ****',
            'User-Agent': 'AzureCLI/2.86.0',
            'x-ms-client-request-id': 'req-1',
            'Content-Type': 'application/octet-stream',
            'Cache-Control': 'no-cache',
        }

        _populate_cached_scm_headers(params, headers)

        # Lowercase key preserved (byte-equivalent to a fresh fetch on this
        # path). User-Agent included. Request id and content-type excluded.
        self.assertEqual(set(params._cached_scm_headers.keys()), {'authorization', 'User-Agent'})
        self.assertEqual(params._cached_scm_headers['authorization'], 'Basic ****')

    def test_populate_cached_scm_headers_aad_capitalized_key(self):
        # The AAD branch of get_scm_site_headers sets headers["Authorization"]
        # (capitalized) for the Bearer token.
        from azure.cli.command_modules.appservice.custom import _populate_cached_scm_headers
        params = self._make_params()
        headers = {
            'Authorization': 'Bearer eyJ0eXAiOiJKV1QiLCJ...',
            'User-Agent': 'AzureCLI/2.86.0',
            'x-ms-client-request-id': 'req-1',
        }

        _populate_cached_scm_headers(params, headers)

        self.assertEqual(set(params._cached_scm_headers.keys()), {'Authorization', 'User-Agent'})
        self.assertEqual(params._cached_scm_headers['Authorization'], 'Bearer eyJ0eXAiOiJKV1QiLCJ...')

    def test_populate_cached_scm_headers_noop_without_authorization(self):
        from azure.cli.command_modules.appservice.custom import _populate_cached_scm_headers
        params = self._make_params()

        _populate_cached_scm_headers(params, {'Content-Type': 'application/octet-stream'})

        self.assertIsNone(params._cached_scm_headers)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers')
    @mock.patch('azure.cli.command_modules.appservice.custom.time.sleep')
    @mock.patch('requests.get')
    def test_check_zip_deployment_status_reuses_cached_headers(
            self, requests_get_mock, _sleep_mock, get_scm_site_headers_mock):
        from azure.cli.command_modules.appservice.custom import _check_zip_deployment_status
        params = self._make_params()
        params._cached_scm_headers = {
            'Authorization': 'Basic ****',
            'User-Agent': 'AzureCLI/test',
        }
        # If the cache is honored, get_scm_site_headers must not be called.
        get_scm_site_headers_mock.side_effect = AssertionError(
            'get_scm_site_headers must not be called when cache is populated')

        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json.return_value = {'status': 4}
        requests_get_mock.return_value = resp

        result = _check_zip_deployment_status(
            params.cmd, 'myRG', 'myApp',
            'https://myapp.scm.azurewebsites.net/api/deployments/latest',
            None, timeout=10, deploy_params=params)

        self.assertEqual(result.get('status'), 4)
        # Auth + UA reused from cache; request id refreshed from cmd.
        sent_headers = requests_get_mock.call_args.kwargs['headers']
        self.assertEqual(sent_headers['Authorization'], 'Basic ****')
        self.assertEqual(sent_headers['User-Agent'], 'AzureCLI/test')
        self.assertEqual(sent_headers['x-ms-client-request-id'], 'req-1')

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers')
    @mock.patch('azure.cli.command_modules.appservice.custom.time.sleep')
    @mock.patch('requests.get')
    def test_check_zip_deployment_status_reuses_cached_headers_basic_auth(
            self, requests_get_mock, _sleep_mock, get_scm_site_headers_mock):
        # Integration-style test: feed the cache via _populate_cached_scm_headers
        # using the exact dict shape urllib3 produces on the basic-auth path
        # (lowercase 'authorization'), then verify the status poller forwards
        # the credential correctly. This is the customer's actual code path.
        from azure.cli.command_modules.appservice.custom import (
            _populate_cached_scm_headers, _check_zip_deployment_status)
        params = self._make_params()
        _populate_cached_scm_headers(params, {
            'authorization': 'Basic ****',  # lowercase from urllib3
            'User-Agent': 'AzureCLI/test',
            'x-ms-client-request-id': 'publish-leg-id',
            'Content-Type': 'application/octet-stream',
        })
        get_scm_site_headers_mock.side_effect = AssertionError(
            'get_scm_site_headers must not be called when cache is populated')

        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json.return_value = {'status': 4}
        requests_get_mock.return_value = resp

        _check_zip_deployment_status(
            params.cmd, 'myRG', 'myApp',
            'https://myapp.scm.azurewebsites.net/api/deployments/latest',
            None, timeout=10, deploy_params=params)

        sent_headers = requests_get_mock.call_args.kwargs['headers']
        # Lowercase key faithfully forwarded — HTTP is case-insensitive so the
        # server treats this the same as 'Authorization'.
        self.assertEqual(sent_headers['authorization'], 'Basic ****')
        self.assertEqual(sent_headers['User-Agent'], 'AzureCLI/test')
        # Fresh request id, not the one from the publish leg.
        self.assertEqual(sent_headers['x-ms-client-request-id'], 'req-1')
        self.assertNotIn('Content-Type', sent_headers)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Bearer fresh', 'User-Agent': 'AzureCLI/test'})
    @mock.patch('azure.cli.command_modules.appservice.custom.time.sleep')
    @mock.patch('requests.get')
    def test_check_zip_deployment_status_falls_back_when_no_cache(
            self, requests_get_mock, _sleep_mock, get_scm_site_headers_mock):
        from azure.cli.command_modules.appservice.custom import _check_zip_deployment_status

        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json.return_value = {'status': 4}
        requests_get_mock.return_value = resp

        # No deploy_params at all — legacy enable_zip_deploy call shape.
        cmd = mock.MagicMock()
        cmd.cli_ctx = mock.MagicMock()
        _check_zip_deployment_status(
            cmd, 'myRG', 'myApp',
            'https://myapp.scm.azurewebsites.net/api/deployments/latest',
            None, timeout=10)

        get_scm_site_headers_mock.assert_called_once_with(cmd.cli_ctx, 'myApp', 'myRG', None)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers',
                return_value={'Authorization': 'Basic ****', 'User-Agent': 'AzureCLI/test'})
    @mock.patch('azure.cli.command_modules.appservice.custom.time.sleep')
    @mock.patch('requests.get')
    def test_check_zip_deployment_status_refreshes_on_401(
            self, requests_get_mock, _sleep_mock, get_scm_site_headers_mock):
        from azure.cli.command_modules.appservice.custom import _check_zip_deployment_status
        params = self._make_params()
        params._cached_scm_headers = {
            'Authorization': 'Basic ****',
            'User-Agent': 'AzureCLI/test',
        }

        resp_401 = mock.MagicMock()
        resp_401.status_code = 401
        resp_ok = mock.MagicMock()
        resp_ok.status_code = 200
        resp_ok.json.return_value = {'status': 4}
        requests_get_mock.side_effect = [resp_401, resp_ok]

        result = _check_zip_deployment_status(
            params.cmd, 'myRG', 'myApp',
            'https://myapp.scm.azurewebsites.net/api/deployments/latest',
            None, timeout=10, deploy_params=params)

        self.assertEqual(result.get('status'), 4)
        # After 401 we refetched once and replaced the cached headers. The
        # is_flex_hint defaults to None when params.is_functionapp is None
        # (test setup); the refresh respects whatever the hint helper returns.
        get_scm_site_headers_mock.assert_called_once_with(
            params.cmd.cli_ctx, 'myApp', 'myRG', None, is_flex_hint=None)
        self.assertEqual(params._cached_scm_headers['Authorization'], 'Basic ****')
        # Second request used the fresh credentials.
        second_call_headers = requests_get_mock.call_args_list[1].kwargs['headers']
        self.assertEqual(second_call_headers['Authorization'], 'Basic ****')

    @mock.patch('azure.cli.command_modules.appservice.custom._make_onedeploy_request')
    @mock.patch('azure.cli.command_modules.appservice.custom._update_artifact_type')
    def test_perform_onedeploy_internal_clears_cache_on_success(
            self, _update_type_mock, make_request_mock):
        from azure.cli.command_modules.appservice.custom import _perform_onedeploy_internal
        params = self._make_params()
        params.enriched_errors = False
        params.is_linux_webapp = False
        params.is_functionapp = False

        def _populate_and_succeed(_params):
            _params._cached_scm_headers = {'Authorization': 'Basic ****'}
            _params._cached_site = mock.MagicMock(name='site')
            return {'status': 'ok'}
        make_request_mock.side_effect = _populate_and_succeed

        result = _perform_onedeploy_internal(params)

        self.assertEqual(result, {'status': 'ok'})
        # finally block must drop all caches even on the happy path.
        self.assertIsNone(params._cached_scm_headers)
        self.assertIsNone(params._cached_site)

    @mock.patch('azure.cli.command_modules.appservice.custom._make_onedeploy_request')
    @mock.patch('azure.cli.command_modules.appservice.custom._update_artifact_type')
    def test_perform_onedeploy_internal_clears_cache_on_exception(
            self, _update_type_mock, make_request_mock):
        from azure.cli.command_modules.appservice.custom import _perform_onedeploy_internal
        params = self._make_params()
        params.enriched_errors = False
        params.is_linux_webapp = False
        params.is_functionapp = False

        def _populate_and_raise(_params):
            _params._cached_scm_headers = {'Authorization': 'Basic ****'}
            _params._cached_site = mock.MagicMock(name='site')
            raise RuntimeError('boom')
        make_request_mock.side_effect = _populate_and_raise

        with self.assertRaises(RuntimeError):
            _perform_onedeploy_internal(params)

        # finally block must drop all caches before the exception propagates,
        # so telemetry / outer handlers cannot see the credential.
        self.assertIsNone(params._cached_scm_headers)
        self.assertIsNone(params._cached_site)

    def test_one_deploy_params_repr_does_not_leak_credentials(self):
        from azure.cli.command_modules.appservice.custom import OneDeployParams
        params = OneDeployParams()
        params._cached_scm_headers = {'Authorization': 'Basic ****'}
        # The default repr is the object id; it must not contain attribute
        # values. If a future change adds a __repr__/__str__ that serializes
        # the cache, this test fails so the reviewer is forced to think about
        # the credential exposure.
        self.assertNotIn('Basic', repr(params))
        self.assertNotIn('c2VjcmV0', repr(params))
        self.assertNotIn('Basic', str(params))

    def test_known_is_flex_hint_web_app(self):
        # Web apps (is_functionapp=False) can never be FlexConsumption, so the
        # hint short-circuits the is_flex_functionapp ARM call.
        from azure.cli.command_modules.appservice.custom import _known_is_flex_hint
        params = self._make_params()
        params.is_functionapp = False
        self.assertEqual(_known_is_flex_hint(params), False)

    def test_known_is_flex_hint_function_app(self):
        # When the cached site has a sku, derive the answer directly.
        from azure.cli.command_modules.appservice.custom import _known_is_flex_hint
        params = self._make_params()
        params.is_functionapp = True
        # No cached site → cannot determine, return None
        params._cached_site = None
        self.assertIsNone(_known_is_flex_hint(params))
        # Cached site with FlexConsumption SKU → True
        params._cached_site = mock.MagicMock(sku='FlexConsumption')
        self.assertTrue(_known_is_flex_hint(params))
        # Cached site with different SKU → False
        params._cached_site = mock.MagicMock(sku='Dynamic')
        self.assertFalse(_known_is_flex_hint(params))

    def test_known_is_flex_hint_unknown(self):
        # Defensive: if is_functionapp hasn't been populated and no cached site,
        # don't pretend to know.
        from azure.cli.command_modules.appservice.custom import _known_is_flex_hint
        params = self._make_params()
        params.is_functionapp = None
        params._cached_site = None
        self.assertIsNone(_known_is_flex_hint(params))
        self.assertIsNone(_known_is_flex_hint(None))

    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp')
    @mock.patch('azure.cli.command_modules.appservice.custom.basic_auth_supported', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._get_site_credential',
                return_value=('user', 'pass'))
    @mock.patch('azure.cli.command_modules.appservice.custom.get_az_user_agent', return_value='AzureCLI/test')
    def test_get_scm_site_headers_skips_is_flex_when_hint_provided(
            self, _ua_mock, _cred_mock, _basic_auth_mock, is_flex_mock):
        # When the caller passes is_flex_hint=False, is_flex_functionapp must
        # not be invoked — saves one ARM call (GET /sites api 2023-12-01).
        from azure.cli.command_modules.appservice.custom import get_scm_site_headers
        cli_ctx = mock.MagicMock()
        cli_ctx.data = {'headers': {'x-ms-client-request-id': 'req-1'}}
        is_flex_mock.side_effect = AssertionError(
            'is_flex_functionapp must not be called when is_flex_hint is provided')

        headers = get_scm_site_headers(cli_ctx, 'myApp', 'myRG', None, is_flex_hint=False)

        # Basic-auth branch was selected because is_flex=False (from hint).
        is_flex_mock.assert_not_called()
        self.assertIn('authorization', headers)  # lowercase from urllib3
        self.assertTrue(headers['authorization'].startswith('Basic '))

    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.basic_auth_supported', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._get_site_credential',
                return_value=('user', 'pass'))
    @mock.patch('azure.cli.command_modules.appservice.custom.get_az_user_agent', return_value='AzureCLI/test')
    def test_get_scm_site_headers_calls_is_flex_when_no_hint(
            self, _ua_mock, _cred_mock, _basic_auth_mock, is_flex_mock):
        # Backward-compat: when no hint is provided (existing callers),
        # is_flex_functionapp is invoked exactly as before.
        from azure.cli.command_modules.appservice.custom import get_scm_site_headers
        cli_ctx = mock.MagicMock()
        cli_ctx.data = {'headers': {'x-ms-client-request-id': 'req-1'}}

        get_scm_site_headers(cli_ctx, 'myApp', 'myRG', None)

        is_flex_mock.assert_called_once_with(cli_ctx, 'myRG', 'myApp')


class TestOneDeploySiteCache(unittest.TestCase):
    """Tests for the per-invocation Site cache on OneDeployParams.

    The Site cache dedupes GET /sites calls between perform_onedeploy_webapp,
    _get_onedeploy_request_body (is_linux check),
    _check_runtimestatus_with_deploymentstatusapi (is_linux check), and the
    success-log URL builder. See _get_or_fetch_site / _get_or_fetch_is_linux_webapp
    / _get_visit_url / _url_from_site in custom.py.
    """

    def _make_params(self):
        from azure.cli.command_modules.appservice.custom import OneDeployParams
        params = OneDeployParams()
        params.cmd = mock.MagicMock()
        params.cmd.cli_ctx = mock.MagicMock()
        params.resource_group_name = 'myRG'
        params.webapp_name = 'myApp'
        params.slot = None
        return params

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    def test_get_or_fetch_site_caches_result(self, generic_op_mock):
        from azure.cli.command_modules.appservice.custom import _get_or_fetch_site
        params = self._make_params()
        sentinel_site = mock.MagicMock(name='site')
        generic_op_mock.return_value = sentinel_site

        first = _get_or_fetch_site(params)
        second = _get_or_fetch_site(params)
        third = _get_or_fetch_site(params)

        self.assertIs(first, sentinel_site)
        self.assertIs(second, sentinel_site)
        self.assertIs(third, sentinel_site)
        generic_op_mock.assert_called_once_with(
            params.cmd.cli_ctx, 'myRG', 'myApp', 'get', None)

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    def test_get_or_fetch_site_uses_slot_when_set(self, generic_op_mock):
        # When params.slot is set, the cached Site must be the slot's Site
        # (via get_slot under the hood), not the production Site. Slots have
        # their own host_name_ssl_states and enabled_host_names, so
        # _get_visit_url would otherwise show the wrong URL.
        from azure.cli.command_modules.appservice.custom import _get_or_fetch_site
        params = self._make_params()
        params.slot = 'staging'
        slot_site = mock.MagicMock(name='slotSite')
        generic_op_mock.return_value = slot_site

        result = _get_or_fetch_site(params)

        self.assertIs(result, slot_site)
        generic_op_mock.assert_called_once_with(
            params.cmd.cli_ctx, 'myRG', 'myApp', 'get', 'staging')

    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    def test_get_or_fetch_is_linux_webapp_uses_cached_value(self, generic_op_mock):
        # perform_onedeploy_webapp populates is_linux_webapp eagerly; the
        # helper must not re-fetch when the answer is already known.
        from azure.cli.command_modules.appservice.custom import _get_or_fetch_is_linux_webapp
        params = self._make_params()
        params.is_linux_webapp = True

        self.assertTrue(_get_or_fetch_is_linux_webapp(params))
        generic_op_mock.assert_not_called()

        params.is_linux_webapp = False
        self.assertFalse(_get_or_fetch_is_linux_webapp(params))
        generic_op_mock.assert_not_called()

    @mock.patch('azure.cli.command_modules.appservice.custom.is_linux_webapp', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    def test_get_or_fetch_is_linux_webapp_lazy_fetch_for_functionapp(
            self, generic_op_mock, is_linux_mock):
        # perform_onedeploy_functionapp does NOT pre-populate is_linux_webapp,
        # so the first consumer must lazily fetch + cache the Site.
        from azure.cli.command_modules.appservice.custom import _get_or_fetch_is_linux_webapp
        params = self._make_params()
        params.is_linux_webapp = None
        site = mock.MagicMock(name='site')
        generic_op_mock.return_value = site

        first = _get_or_fetch_is_linux_webapp(params)
        second = _get_or_fetch_is_linux_webapp(params)

        self.assertTrue(first)
        self.assertTrue(second)
        # Site fetched exactly once even across multiple lookups.
        generic_op_mock.assert_called_once()
        is_linux_mock.assert_called_once_with(site)
        # Result is memoized on params for subsequent helpers (e.g. the
        # status poller).
        self.assertTrue(params.is_linux_webapp)
        self.assertIs(params._cached_site, site)

    def test_url_from_site_picks_https_when_ssl_enabled(self):
        from azure.cli.command_modules.appservice.custom import _url_from_site
        SslState = mock.MagicMock()
        SslState.disabled = 'Disabled'
        cmd = mock.MagicMock()
        cmd.get_models.return_value = SslState
        site = mock.MagicMock()
        site.enabled_host_names = ['custom.contoso.com', 'myapp.azurewebsites.net']
        site.host_name_ssl_states = [
            mock.MagicMock(ssl_state='SniEnabled'),
        ]

        self.assertEqual(
            _url_from_site(cmd, site), 'https://custom.contoso.com')

    def test_url_from_site_picks_http_when_no_ssl(self):
        from azure.cli.command_modules.appservice.custom import _url_from_site
        SslState = mock.MagicMock()
        SslState.disabled = 'Disabled'
        cmd = mock.MagicMock()
        cmd.get_models.return_value = SslState
        site = mock.MagicMock()
        site.enabled_host_names = ['myapp.azurewebsites.net']
        site.host_name_ssl_states = [
            mock.MagicMock(ssl_state='Disabled'),
            mock.MagicMock(ssl_state='Disabled'),
        ]

        self.assertEqual(
            _url_from_site(cmd, site), 'http://myapp.azurewebsites.net')

    @mock.patch('azure.cli.command_modules.appservice.custom._get_url')
    @mock.patch('azure.cli.command_modules.appservice.custom._url_from_site',
                return_value='https://myapp.azurewebsites.net')
    def test_get_visit_url_uses_cached_site(self, url_from_site_mock, get_url_mock):
        # When the Site is already cached (the common case after
        # perform_onedeploy_webapp), no fallback ARM call should be made.
        from azure.cli.command_modules.appservice.custom import _get_visit_url
        params = self._make_params()
        params._cached_site = mock.MagicMock(name='site')

        result = _get_visit_url(params)

        self.assertEqual(result, 'https://myapp.azurewebsites.net')
        url_from_site_mock.assert_called_once_with(params.cmd, params._cached_site)
        get_url_mock.assert_not_called()

    @mock.patch('azure.cli.command_modules.appservice.custom._get_url')
    @mock.patch('azure.cli.command_modules.appservice.custom._url_from_site',
                return_value='https://myapp-staging.azurewebsites.net')
    def test_get_visit_url_uses_cached_slot_site(self, url_from_site_mock, get_url_mock):
        # For slot deployments, perform_onedeploy_webapp fetches the slot's
        # Site, so _get_visit_url can still serve the URL from the cache.
        from azure.cli.command_modules.appservice.custom import _get_visit_url
        params = self._make_params()
        params.slot = 'staging'
        params._cached_site = mock.MagicMock(name='slotSite')

        result = _get_visit_url(params)

        self.assertEqual(result, 'https://myapp-staging.azurewebsites.net')
        url_from_site_mock.assert_called_once_with(params.cmd, params._cached_site)
        get_url_mock.assert_not_called()

    @mock.patch('azure.cli.command_modules.appservice.custom._get_url',
                return_value='https://myapp.azurewebsites.net')
    def test_get_visit_url_falls_back_when_no_cache(self, get_url_mock):
        # perform_onedeploy_functionapp does not populate _cached_site; in
        # that path we must still produce a URL via the standard helper.
        from azure.cli.command_modules.appservice.custom import _get_visit_url
        params = self._make_params()
        params._cached_site = None
        params.slot = None

        result = _get_visit_url(params)

        self.assertEqual(result, 'https://myapp.azurewebsites.net')
        get_url_mock.assert_called_once_with(params.cmd, 'myRG', 'myApp', None)


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import unittest
from unittest import mock
import os

from azure.mgmt.web import WebSiteManagementClient
from knack.util import CLIError
from azure.cli.command_modules.appservice.custom import (
    enable_zip_deploy_functionapp,
    enable_zip_deploy,
    enable_zip_deploy_flex,
    check_flex_app_after_deployment,
    add_remote_build_app_settings,
    remove_remote_build_app_settings,
    config_source_control,
    validate_app_settings_in_scm,
    update_container_settings_functionapp,
    list_function_keys)
from azure.cli.core.profiles import ResourceType
from azure.cli.core.azclierror import (AzureInternalError, UnclassifiedUserFault)
from azure.cli.core.azclierror import ResourceNotFoundError

TEST_DIR = os.path.abspath(os.path.join(os.path.abspath(__file__), '..'))


def _get_test_cmd():
    from azure.cli.core.mock import DummyCli
    from azure.cli.core import AzCommandsLoader
    from azure.cli.core.commands import AzCliCommand
    cli_ctx = DummyCli()
    loader = AzCommandsLoader(cli_ctx, resource_type=ResourceType.MGMT_APPSERVICE)
    cmd = AzCliCommand(loader, 'test', None)
    cmd.command_kwargs = {'resource_type': ResourceType.MGMT_APPSERVICE}
    cmd.cli_ctx = cli_ctx
    return cmd


def _get_zip_deploy_headers(username, password, cmd_mock_client):
    from urllib3.util import make_headers
    from azure.cli.core.util import get_az_user_agent

    headers = make_headers(basic_auth='{0}:{1}'.format(username, password))
    headers['Content-Type'] = 'application/octet-stream'
    headers['Cache-Control'] = 'no-cache'
    headers['User-Agent'] = get_az_user_agent()
    headers['x-ms-client-request-id'] = cmd_mock_client.data['headers']['x-ms-client-request-id']
    return headers


def _get_flex_zip_deploy_headers(cmd_mock_client):
    from urllib3.util import make_headers
    from azure.cli.core.util import get_az_user_agent

    headers = make_headers()
    headers['Authorization'] = "Bearer 1234"
    headers['Content-Type'] = 'application/zip'
    headers['Cache-Control'] = 'no-cache'
    headers['User-Agent'] = get_az_user_agent()
    headers['x-ms-client-request-id'] = cmd_mock_client.data['headers']['x-ms-client-request-id']
    return headers


class TestFunctionappMocked(unittest.TestCase):
    def setUp(self):
        self.client = WebSiteManagementClient(mock.MagicMock(), '123455678')

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.parse_resource_id')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.enable_zip_deploy')
    @mock.patch('azure.cli.command_modules.appservice.custom.add_remote_build_app_settings')
    def test_functionapp_zip_deploy_flow(self,
                                         add_remote_build_app_settings_mock,
                                         enable_zip_deploy_mock,
                                         is_flex_functionapp_mock,
                                         parse_resource_id_mock,
                                         web_client_factory_mock):
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        # action
        enable_zip_deploy_functionapp(cmd_mock, 'rg', 'name', 'src', build_remote=True, timeout=None, slot=None)

        # assert
        parse_resource_id_mock.assert_called_once()
        enable_zip_deploy_mock.assert_called_once()
        add_remote_build_app_settings_mock.assert_called_once()

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.parse_resource_id')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.enable_zip_deploy')
    @mock.patch('azure.cli.command_modules.appservice.custom.remove_remote_build_app_settings')
    def test_functionapp_zip_deploy_flow(self,
                                         remove_remote_build_app_settings_mock,
                                         enable_zip_deploy_mock,
                                         is_flex_functionapp_mock,
                                         parse_resource_id_mock,
                                         web_client_factory_mock):
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        # action
        enable_zip_deploy_functionapp(cmd_mock, 'rg', 'name', 'src', build_remote=False, timeout=None, slot=None)

        # assert
        parse_resource_id_mock.assert_called_once()
        enable_zip_deploy_mock.assert_called_once()
        remove_remote_build_app_settings_mock.assert_called_once()

    @mock.patch('azure.cli.command_modules.appservice.custom.check_language_runtime')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.parse_resource_id')
    @mock.patch('azure.cli.command_modules.appservice.custom.validate_zip_deploy_app_setting_exists')
    @mock.patch('azure.cli.command_modules.appservice.custom.upload_zip_to_storage')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_plan_consumption', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', return_value=False)
    def test_functionapp_linux_consumption_non_remote_build(self,
                                                            is_flex_functionapp_mock,
                                                            is_plan_consumption_mock,
                                                            upload_zip_to_storage_mock,
                                                            validate_zip_deploy_app_setting_exists_mock,
                                                            parse_resource_id_mock,
                                                            web_client_factory_mock,
                                                            check_language_runtime_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        appservice_mock = mock.Mock()
        appservice_mock.reserved = True  # Marked app service as Linux

        web_client_mock = mock.Mock()
        web_client_mock.web_apps = mock.Mock()
        web_client_mock.web_apps.get = mock.Mock(return_value=appservice_mock)
        web_client_factory_mock.return_value = web_client_mock

        # action
        # Linux Consumption app should use update-storage to deploy when not using remote build
        enable_zip_deploy_functionapp(cmd_mock, 'rg', 'name', 'src', build_remote=False, timeout=None, slot=None)

        # assert
        web_client_mock.web_apps.get.assert_called_with('rg', 'name')
        upload_zip_to_storage_mock.assert_called_with(cmd_mock, 'rg', 'name', 'src', None)

    @mock.patch('azure.cli.command_modules.appservice.custom.check_language_runtime')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.parse_resource_id')
    @mock.patch('azure.cli.command_modules.appservice.custom.validate_zip_deploy_app_setting_exists')
    @mock.patch('azure.cli.command_modules.appservice.custom.upload_zip_to_storage')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_plan_consumption', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', return_value=False)
    def test_functionapp_linux_consumption_non_remote_build_with_slot(self,
                                                            is_flex_functionapp_mock,
                                                            is_plan_consumption_mock,
                                                            upload_zip_to_storage_mock,
                                                            validate_zip_deploy_app_setting_exists_mock,
                                                            parse_resource_id_mock,
                                                            web_client_factory_mock,
                                                            check_language_runtime_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        appservice_mock = mock.Mock()
        appservice_mock.reserved = True  # Marked app service as Linux

        web_client_mock = mock.Mock()
        web_client_mock.web_apps = mock.Mock()
        web_client_mock.web_apps.get = mock.Mock(return_value=appservice_mock)
        web_client_factory_mock.return_value = web_client_mock

        # action
        # Linux Consumption app should use update-storage to deploy when not using remote build
        enable_zip_deploy_functionapp(cmd_mock, 'rg', 'name', 'src', build_remote=False, timeout=None, slot='slot')

        # assert
        web_client_mock.web_apps.get.assert_called_with('rg', 'name')
        upload_zip_to_storage_mock.assert_called_with(cmd_mock, 'rg', 'name', 'src', 'slot')

    @mock.patch('azure.cli.command_modules.appservice.custom.check_language_runtime')
    @mock.patch('azure.cli.command_modules.appservice.custom.add_remote_build_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.parse_resource_id')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_flex_functionapp', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.enable_zip_deploy')
    def test_functionapp_remote_build_supports_linux(self,
                                                     enable_zip_deploy_mock,
                                                     is_flex_functionapp_mock,
                                                     parse_resource_id_mock,
                                                     web_client_factory_mock,
                                                     add_remote_build_app_settings_mock,
                                                     check_language_runtime_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        appservice_mock = mock.Mock()
        appservice_mock.reserved = True  # Marked app service as Linux

        web_client_mock = mock.Mock()
        web_client_mock.web_apps = mock.Mock()
        web_client_mock.web_apps.get = mock.Mock(return_value=appservice_mock)
        web_client_factory_mock.return_value = web_client_mock

        # action
        enable_zip_deploy_functionapp(cmd_mock, 'rg', 'name', 'src', build_remote=True, timeout=None, slot=None)

        # assert
        web_client_mock.web_apps.get.assert_called_with('rg', 'name')
        enable_zip_deploy_mock.assert_called_with(cmd_mock, 'rg', 'name', 'src', None, None)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', return_value='https://mock-scm')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_runtime_config')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers_flex')
    @mock.patch('requests.post', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._check_zip_deployment_status_flex')
    def test_enable_zip_deploy_flex(self,
                                    check_zip_deployment_status_mock,
                                    requests_post_mock,
                                    get_scm_site_headers_flex_mock,
                                    get_runtime_config_mock,
                                    get_scm_url_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        response = mock.MagicMock()
        response.status_code = 202
        requests_post_mock.return_value = response

        get_runtime_config_mock.return_value = {
            "name": "java",
        }

        expected_zip_deploy_headers = _get_flex_zip_deploy_headers(cmd_mock.cli_ctx)
        get_scm_site_headers_flex_mock.return_value = expected_zip_deploy_headers

        # action
        with mock.patch('builtins.open', new_callable=mock.mock_open, read_data='zip-content'):
            enable_zip_deploy_flex(cmd_mock, 'rg', 'name', 'src', slot=None, build_remote=True)

        # assert
        requests_post_mock.assert_called_with('https://mock-scm/api/publish?RemoteBuild=True&Deployer=az_cli', data='zip-content',
                                              headers=expected_zip_deploy_headers, verify=mock.ANY)
        # TODO improve authorization matcher
        check_zip_deployment_status_mock.assert_called_with(cmd_mock, 'rg', 'name',
                                                            'https://mock-scm/api/deployments/latest', None)

    @mock.patch('time.sleep')
    @mock.patch('requests.get', autospec=True)
    @mock.patch('azure.cli.core.util.should_disable_connection_verify', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.list_host_keys')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_host_url',
                return_value='https://mock-func.azurewebsites.net')
    def test_check_flex_app_after_deployment_success(self,
                                                     get_host_url_mock,
                                                     list_host_keys_mock,
                                                     should_disable_verify_mock,
                                                     requests_get_mock,
                                                     sleep_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        list_host_keys_mock.return_value = mock.Mock(master_key='master-key')
        response = mock.Mock(status_code=200, reason='OK')
        requests_get_mock.return_value = response

        # action
        result = check_flex_app_after_deployment(cmd_mock, 'rg', 'name')

        # assert
        self.assertEqual(result, "Deployment was successful.")
        requests_get_mock.assert_called_with('https://mock-func.azurewebsites.net/admin/host/status',
                                             headers={"x-functions-key": 'master-key'},
                                             verify=True)

    @mock.patch('time.sleep')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_host_url', side_effect=ValueError())
    def test_check_flex_app_after_deployment_host_url_fetch_failure(self,
                                                                    get_host_url_mock,
                                                                    sleep_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        with self.assertRaises(ResourceNotFoundError):
            check_flex_app_after_deployment(cmd_mock, 'rg', 'name')

        # assert
        get_host_url_mock.assert_called_once_with(cmd_mock, 'rg', 'name')

    @mock.patch('time.sleep')
    @mock.patch('azure.cli.command_modules.appservice.custom.list_host_keys', side_effect=Exception())
    @mock.patch('azure.cli.command_modules.appservice.custom._get_host_url',
                return_value='https://mock-func.azurewebsites.net')
    def test_check_flex_app_after_deployment_host_key_fetch_failure(self,
                                                                    get_host_url_mock,
                                                                    list_host_keys_mock,
                                                                    sleep_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        with self.assertRaises(ResourceNotFoundError):
            check_flex_app_after_deployment(cmd_mock, 'rg', 'name')

        # assert
        list_host_keys_mock.assert_called_once_with(cmd_mock, 'rg', 'name')

    @mock.patch('time.sleep')
    @mock.patch('requests.get', autospec=True)
    @mock.patch('azure.cli.core.util.should_disable_connection_verify', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.list_host_keys')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_host_url',
                return_value='https://mock-func.azurewebsites.net')
    def test_check_flex_app_after_deployment_ip_restriction(self,
                                                            get_host_url_mock,
                                                            list_host_keys_mock,
                                                            should_disable_verify_mock,
                                                            requests_get_mock,
                                                            sleep_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        list_host_keys_mock.return_value = mock.Mock(master_key='master-key')
        requests_get_mock.return_value = mock.Mock(status_code=403, reason='Ip Forbidden')

        # action
        result = check_flex_app_after_deployment(cmd_mock, 'rg', 'name')

        # assert
        self.assertEqual(result, "Deployment was successful but health check failed due to IP restriction.")

    @mock.patch('time.sleep')
    @mock.patch('requests.get', autospec=True)
    @mock.patch('azure.cli.core.util.should_disable_connection_verify', return_value=False)
    @mock.patch('azure.cli.command_modules.appservice.custom.list_host_keys')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_host_url',
                return_value='https://mock-func.azurewebsites.net')
    def test_check_flex_app_after_deployment_unhealthy(self,
                                                       get_host_url_mock,
                                                       list_host_keys_mock,
                                                       should_disable_verify_mock,
                                                       requests_get_mock,
                                                       sleep_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        list_host_keys_mock.return_value = mock.Mock(master_key='master-key')
        requests_get_mock.return_value = mock.Mock(status_code=500, reason='Internal Server Error')

        # action
        with self.assertRaises(CLIError):
            check_flex_app_after_deployment(cmd_mock, 'rg', 'name')

        # assert
        self.assertEqual(requests_get_mock.call_count, 15)


    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', side_effect=ValueError())
    def test_enable_zip_deploy_remote_build_no_scm_site(self,
                                                        get_scm_url_mock,
                                                        get_scm_headers_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        # action
        # When the function app is created before 8/1/2019, it cannot use remote build
        with self.assertRaises(CLIError):
            enable_zip_deploy(cmd_mock, 'rg', 'name', 'src', slot=None)

        # assert
        get_scm_url_mock.assert_called_with(cmd_mock, 'rg', 'name', None)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', return_value='https://mock-scm')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('requests.post', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom._check_zip_deployment_status')
    def test_enable_zip_deploy_accepted(self,
                                        check_zip_deployment_status_mock,
                                        requests_post_mock,
                                        web_client_factory_mock,
                                        get_scm_url_mock,
                                        get_scm_headers_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        response = mock.MagicMock()
        response.status_code = 202
        requests_post_mock.return_value = response

        appservice_mock = mock.Mock()
        appservice_mock.kind = "functionapp"

        web_client_mock = mock.Mock()
        web_client_mock.web_apps = mock.Mock()
        web_client_mock.web_apps.get = mock.Mock(return_value=appservice_mock)
        web_client_factory_mock.return_value = web_client_mock

        expected_zip_deploy_headers = _get_zip_deploy_headers('usr', 'pwd', cmd_mock.cli_ctx)
        get_scm_headers_mock.return_value = expected_zip_deploy_headers

        # action
        with mock.patch('builtins.open', new_callable=mock.mock_open, read_data='zip-content'):
            enable_zip_deploy(cmd_mock, 'rg', 'name', 'src', slot=None)

        # assert
        requests_post_mock.assert_called_with('https://mock-scm/api/zipdeploy?isAsync=true&Deployer=az_cli_functions', data='zip-content',
                                              headers=expected_zip_deploy_headers, verify=mock.ANY)
        # TODO improve authorization matcher
        check_zip_deployment_status_mock.assert_called_with(cmd_mock, 'rg', 'name',
                                                            'https://mock-scm/api/deployments/latest', mock.ANY, None)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', return_value='https://mock-scm')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('requests.post', autospec=True)
    def test_enable_zip_deploy_conflict(self,
                                        requests_post_mock,
                                        web_client_factory_mock,
                                        get_scm_url_mock,
                                        get_scm_headers_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        response = mock.MagicMock()
        response.status_code = 409
        requests_post_mock.return_value = response

        appservice_mock = mock.Mock()
        appservice_mock.kind = "functionapp"

        web_client_mock = mock.Mock()
        web_client_mock.web_apps = mock.Mock()
        web_client_mock.web_apps.get = mock.Mock(return_value=appservice_mock)
        web_client_factory_mock.return_value = web_client_mock

        expected_zip_deploy_headers = _get_zip_deploy_headers('usr', 'pwd', cmd_mock.cli_ctx)
        get_scm_headers_mock.return_value = expected_zip_deploy_headers

        # action
        with mock.patch('builtins.open', new_callable=mock.mock_open, read_data='zip-content'):
            with self.assertRaises(UnclassifiedUserFault):
                enable_zip_deploy(cmd_mock, 'rg', 'name', 'src', slot=None)

        # assert
        requests_post_mock.assert_called_with('https://mock-scm/api/zipdeploy?isAsync=true&Deployer=az_cli_functions', data='zip-content',
                                              headers=expected_zip_deploy_headers, verify=mock.ANY)

    @mock.patch('azure.cli.command_modules.appservice.custom.get_scm_site_headers')
    @mock.patch('azure.cli.command_modules.appservice.custom._get_scm_url', return_value='https://mock-scm')
    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('requests.post', autospec=True)
    def test_enable_zip_deploy_service_unavailable(self,
                                                   requests_post_mock,
                                                   web_client_factory_mock,
                                                   get_scm_url_mock,
                                                   get_scm_headers_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        response = mock.MagicMock()
        response.status_code = 503
        requests_post_mock.return_value = response

        appservice_mock = mock.Mock()
        appservice_mock.kind = "functionapp"

        web_client_mock = mock.Mock()
        web_client_mock.web_apps = mock.Mock()
        web_client_mock.web_apps.get = mock.Mock(return_value=appservice_mock)
        web_client_factory_mock.return_value = web_client_mock


        expected_zip_deploy_headers = _get_zip_deploy_headers('usr', 'pwd', cmd_mock.cli_ctx)
        get_scm_headers_mock.return_value = expected_zip_deploy_headers

        # action
        with mock.patch('builtins.open', new_callable=mock.mock_open, read_data='zip-content'):
            with self.assertRaises(AzureInternalError):
                enable_zip_deploy(cmd_mock, 'rg', 'name', 'src', slot=None)

        # assert
        requests_post_mock.assert_called_with('https://mock-scm/api/zipdeploy?isAsync=true&Deployer=az_cli_functions', data='zip-content',
                                              headers=expected_zip_deploy_headers, verify=mock.ANY)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_settings_from_scm', return_value={
        'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'
    })
    def test_validate_app_settings_in_scm_should_have(self, get_app_settings_from_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        should_have = ['SCM_DO_BUILD_DURING_DEPLOYMENT']

        # action
        result = validate_app_settings_in_scm(cmd_mock, 'rg', 'name', slot=None, should_have=should_have)

        # assert
        self.assertTrue(result)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_settings_from_scm', return_value={
        'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'
    })
    def test_validate_app_settings_in_scm_should_not_have(self, get_app_settings_from_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        should_not_have = ['ENABLE_ORYX_BUILD']

        # action
        result = validate_app_settings_in_scm(cmd_mock, 'rg', 'name', slot=None, should_not_have=should_not_have)

        # assert
        self.assertTrue(result)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_settings_from_scm', return_value={
        'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'
    })
    def test_validate_app_settings_in_scm_should_contain(self, get_app_settings_from_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        should_contain = {'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'}

        # action
        result = validate_app_settings_in_scm(cmd_mock, 'rg', 'name', slot=None, should_contain=should_contain)

        # assert
        self.assertTrue(result)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_settings_from_scm', return_value={
        'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'
    })
    def test_validate_app_settings_in_scm_should_have_failure(self, get_app_settings_from_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        should_have = ['ENABLE_ORYX_BUILD']

        # action
        result = validate_app_settings_in_scm(cmd_mock, 'rg', 'name', slot=None, should_have=should_have)

        # assert
        self.assertFalse(result)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_settings_from_scm', return_value={
        'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'
    })
    def test_validate_app_Settings_in_scm_should_not_have_failure(self, get_app_settings_from_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        should_not_have = ['SCM_DO_BUILD_DURING_DEPLOYMENT']

        # action
        result = validate_app_settings_in_scm(cmd_mock, 'rg', 'name', slot=None, should_not_have=should_not_have)

        # assert
        self.assertFalse(result)

    @mock.patch('azure.cli.command_modules.appservice.custom._get_app_settings_from_scm', return_value={
        'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'
    })
    def test_validate_app_settings_in_scm_should_contain_failure(self, get_app_settings_from_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        should_contain = {'SCM_DO_BUILD_DURING_DEPLOYMENT': 'false'}

        # action
        result = validate_app_settings_in_scm(cmd_mock, 'rg', 'name', slot=None, should_contain=should_contain)

        # assert
        self.assertFalse(result)

    @mock.patch('azure.cli.command_modules.appservice.custom.check_language_runtime')
    @mock.patch('azure.cli.command_modules.appservice.custom.is_centauri_functionapp')
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation')
    @mock.patch('azure.cli.command_modules.appservice.custom.update_functionapp_polling', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_container_settings', autospec=True)
    def test_update_container_settings_functionapp(self,
                                                   update_container_settings_mock,
                                                   update_functionapp_polling_mock,
                                                   site_op_mock,
                                                   is_centauri_functionapp_mock,
                                                   check_language_runtime_mock):
        # prepare
        cmd_mock = _get_test_cmd()
        cli_ctx_mock = mock.MagicMock()
        cmd_mock.cli_ctx = cli_ctx_mock

        Site, DaprConfig, ResourceConfig = cmd_mock.get_models('Site', 'DaprConfig', 'ResourceConfig')
        site = Site(dapr_config=None, location='westus', name='name', resource_config=ResourceConfig())
        site_op_mock.return_value = site

        is_centauri_functionapp_mock.return_value = True

        check_language_runtime_mock.return_value = True

        # action
        update_container_settings_functionapp(cmd_mock, 'rg', 'name', workload_profile_name='d4', cpu=0.5, memory='1Gi')

        # assert
        updated_site = site
        updated_site.dapr_config = DaprConfig()
        update_functionapp_polling_mock.assert_called_with(cmd_mock, 'rg', 'name', updated_site)


    @mock.patch('azure.cli.command_modules.appservice.custom.validate_app_settings_in_scm', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.delete_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_settings', return_value=[])
    def test_add_remote_build_app_settings_add_scm_do_build_during_deployment(self,
                                                                              get_app_settings_mock,
                                                                              delete_app_settings_mock,
                                                                              update_app_settings_mock,
                                                                              validate_app_settings_in_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        add_remote_build_app_settings(cmd_mock, 'rg', 'name', slot=None)

        # assert
        update_app_settings_mock.assert_called_with(cmd_mock, 'rg', 'name', ['SCM_DO_BUILD_DURING_DEPLOYMENT=true'], None)
        validate_app_settings_in_scm_mock.assert_called_with(cmd_mock, 'rg', 'name', None,
                                                             should_contain={'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'},
                                                             should_not_have=[])

    @mock.patch('azure.cli.command_modules.appservice.custom.validate_app_settings_in_scm',
                return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.delete_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_settings', return_value=[
        {
            'name': 'WEBSITE_RUN_FROM_PACKAGE',
            'value': 'https://microsoft.com'
        },
        {
            'name': 'ENABLE_ORYX_BUILD',
            'value': 'true'
        }
    ])
    def test_add_remote_build_app_settings_remove_unnecessary_app_settings(self,
                                                                           get_app_settings_mock,
                                                                           delete_app_settings_mock,
                                                                           update_app_settings_mock,
                                                                           validate_app_settings_in_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        add_remote_build_app_settings(cmd_mock, 'rg', 'name', slot=None)

        # assert
        delete_app_settings_mock.assert_any_call(cmd_mock, 'rg', 'name', ['WEBSITE_RUN_FROM_PACKAGE'], None)
        delete_app_settings_mock.assert_any_call(cmd_mock, 'rg', 'name', ['ENABLE_ORYX_BUILD'], None)
        validate_app_settings_in_scm_mock.assert_called_with(cmd_mock, 'rg', 'name', None,
                                                             should_contain={'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'},
                                                             should_not_have=['WEBSITE_RUN_FROM_PACKAGE', 'ENABLE_ORYX_BUILD'])

    @mock.patch('azure.cli.command_modules.appservice.custom.validate_app_settings_in_scm', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.delete_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_settings', return_value=[{
        'name': 'SCM_DO_BUILD_DURING_DEPLOYMENT',
        'value': 'false'
    }])
    def test_add_remote_build_app_settings_change_scm_do_build_during_deployment(self,
                                                                                 get_app_settings_mock,
                                                                                 delete_app_settings_mock,
                                                                                 update_app_settings_mock,
                                                                                 validate_app_settings_in_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        add_remote_build_app_settings(cmd_mock, 'rg', 'name', slot=None)

        # assert
        update_app_settings_mock.assert_called_with(cmd_mock, 'rg', 'name', ['SCM_DO_BUILD_DURING_DEPLOYMENT=true'], None)
        validate_app_settings_in_scm_mock.assert_called_with(cmd_mock, 'rg', 'name', None,
                                                             should_contain={'SCM_DO_BUILD_DURING_DEPLOYMENT': 'true'},
                                                             should_not_have=[])

    @mock.patch('azure.cli.command_modules.appservice.custom.validate_app_settings_in_scm', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.delete_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_settings', return_value=[{
        'name': 'SCM_DO_BUILD_DURING_DEPLOYMENT',
        'value': 'true'
    }])
    def test_add_remote_build_app_settings_do_nothing(self,
                                                      get_app_settings_mock,
                                                      delete_app_settings_mock,
                                                      update_app_settings_mock,
                                                      validate_app_settings_in_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        add_remote_build_app_settings(cmd_mock, 'rg', 'name', slot=None)

        # assert
        update_app_settings_mock.assert_not_called()
        validate_app_settings_in_scm_mock.assert_not_called()

    @mock.patch('azure.cli.command_modules.appservice.custom.validate_app_settings_in_scm', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.delete_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_settings', return_value=[])
    def test_remove_remote_build_app_settings_disable_scm_do_build_during_deployment(self,
                                                                                     get_app_settings_mock,
                                                                                     delete_app_settings_mock,
                                                                                     update_app_settings_mock,
                                                                                     validate_app_settings_in_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        remove_remote_build_app_settings(cmd_mock, 'rg', 'name', slot=None)

        # assert
        update_app_settings_mock.assert_called_with(cmd_mock, 'rg', 'name', ['SCM_DO_BUILD_DURING_DEPLOYMENT=false'], None)
        validate_app_settings_in_scm_mock.assert_called_with(cmd_mock, 'rg', 'name', None,
                                                             should_contain={'SCM_DO_BUILD_DURING_DEPLOYMENT': 'false'})

    @mock.patch('azure.cli.command_modules.appservice.custom.validate_app_settings_in_scm', return_value=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.update_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.delete_app_settings')
    @mock.patch('azure.cli.command_modules.appservice.custom.get_app_settings', return_value=[{
        'name': 'SCM_DO_BUILD_DURING_DEPLOYMENT',
        'value': 'false'
    }])
    def test_remove_remote_build_app_settings_do_nothing(self,
                                                         get_app_settings_mock,
                                                         delete_app_settings_mock,
                                                         update_app_settings_mock,
                                                         validate_app_settings_in_scm_mock):
        # prepare
        cmd_mock = _get_test_cmd()

        # action
        remove_remote_build_app_settings(cmd_mock, 'rg', 'name', slot=None)

        # assert
        update_app_settings_mock.assert_not_called()
        validate_app_settings_in_scm_mock.assert_not_called()


    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.utils._get_location_from_webapp')
    @mock.patch('azure.cli.command_modules.appservice.custom._generic_site_operation', autospec=True)
    @mock.patch('azure.cli.command_modules.appservice.custom.LongRunningOperation.__call__', autospec=True)
    def test_config_source_control(self,
                                   long_running_operation_mock,
                                   site_op_mock,
                                   location_mock,
                                   web_client_factory_mock):
        # prepare
        client = mock.Mock()
        web_client_factory_mock.return_value = client

        location_mock.return_value = mock.MagicMock()

        site_op_mock.return_value = mock.MagicMock()

        cmd_mock = _get_test_cmd()

        SiteSourceControl, GitHubActionConfiguration, GitHubActionContainerConfiguration = cmd_mock.get_models('SiteSourceControl', 'GitHubActionConfiguration', 'GitHubActionContainerConfiguration')
        container_config = GitHubActionContainerConfiguration(username="username", password="password")
        github_action_config = GitHubActionConfiguration(container_configuration=container_config)
        source_control = SiteSourceControl(git_hub_action_configuration=github_action_config)

        long_running_operation_mock.return_value = source_control

        # action
        response = config_source_control(cmd_mock, 'rg', 'functionapp', 'https://github.com/yugang/azure-site-test')

        # assert
        self.assertEqual(response.git_hub_action_configuration.container_configuration.password, None)

    def test_flex_parse_raw_stacks_handles_empty_app_settings_dictionary(self):
        from azure.cli.command_modules.appservice.custom import _FlexFunctionAppStackRuntimeHelper

        # prepare
        cmd_mock = _get_test_cmd()
        go_stack = {
            'name': 'go',
            'properties': {'majorVersions': [{'minorVersions': [{
                'value': '1.0',
                'stackSettings': {'linuxRuntimeSettings': {
                    'appSettingsDictionary': {},
                    'Sku': [{'skuCode': 'FC1'}],
                    'gitHubActionSettings': {'isSupported': True},
                    'appInsightsSettings': {'isSupported': True},
                    'isDefault': True,
                }},
            }]}]},
        }

        # action
        with mock.patch.object(_FlexFunctionAppStackRuntimeHelper,
                               'get_flex_raw_function_app_stacks',
                               return_value=[go_stack]):
            matched = _FlexFunctionAppStackRuntimeHelper(cmd_mock, 'westcentralus', 'go').resolve('go', '1.0')

        # assert
        self.assertEqual(matched.name, 'go')
        self.assertEqual(matched.version, '1.0')

    def test_flex_parse_raw_stacks_prefers_functions_worker_runtime_when_present(self):
        # FUNCTIONS_WORKER_RUNTIME overrides runtime['name'] when set (e.g., dotnet-isolated under the dotnet stack).
        from azure.cli.command_modules.appservice.custom import _FlexFunctionAppStackRuntimeHelper

        # prepare
        cmd_mock = _get_test_cmd()
        dotnet_isolated_stack = {
            'name': 'dotnet',
            'properties': {'majorVersions': [{'minorVersions': [{
                'value': '8.0',
                'stackSettings': {'linuxRuntimeSettings': {
                    'appSettingsDictionary': {'FUNCTIONS_WORKER_RUNTIME': 'dotnet-isolated'},
                    'Sku': [{'skuCode': 'FC1'}],
                    'gitHubActionSettings': {'isSupported': True},
                    'appInsightsSettings': {'isSupported': True},
                    'isDefault': True,
                }},
            }]}]},
        }

        # action
        with mock.patch.object(_FlexFunctionAppStackRuntimeHelper,
                               'get_flex_raw_function_app_stacks',
                               return_value=[dotnet_isolated_stack]):
            matched = _FlexFunctionAppStackRuntimeHelper(
                cmd_mock, 'westcentralus', 'dotnet-isolated').resolve('dotnet-isolated', '8.0')

        # assert
        self.assertEqual(matched.name, 'dotnet-isolated')

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    def test_list_function_keys_unwraps_broken_string_dictionary(self, web_client_factory_mock):
        # azure-mgmt-web 11.0.0 deserializes flat dictionary responses with .properties = None
        from azure.mgmt.web import models as _models
        from azure.mgmt.web._utils.model_base import _deserialize

        broken = _deserialize(
            _models.StringDictionary,
            {'default': 'vvrX4LJY1JWbimFI28UM', 'myCustomKey': 'abc'})
        self.assertIsNone(broken.properties)

        cmd_mock = _get_test_cmd()
        client_mock = mock.MagicMock()
        client_mock.web_apps.list_function_keys.return_value = broken
        web_client_factory_mock.return_value = client_mock

        result = list_function_keys(cmd_mock, 'rg', 'app', 'httpget')

        self.assertEqual(result, {'default': 'vvrX4LJY1JWbimFI28UM', 'myCustomKey': 'abc'})
        client_mock.web_apps.list_function_keys.assert_called_once_with('rg', 'app', 'httpget')
        client_mock.web_apps.list_function_keys_slot.assert_not_called()

    @mock.patch('azure.cli.command_modules.appservice.custom.web_client_factory', autospec=True)
    def test_list_function_keys_uses_properties_when_sdk_returns_enveloped_response(self, web_client_factory_mock):
        # Prefers .properties when populated (forward-compatible with a future SDK fix)
        fixed = mock.MagicMock()
        fixed.properties = {'default': 'abc'}

        cmd_mock = _get_test_cmd()
        client_mock = mock.MagicMock()
        client_mock.web_apps.list_function_keys_slot.return_value = fixed
        web_client_factory_mock.return_value = client_mock

        result = list_function_keys(cmd_mock, 'rg', 'app', 'httpget', slot='staging')

        self.assertEqual(result, {'default': 'abc'})
        client_mock.web_apps.list_function_keys_slot.assert_called_once_with('rg', 'app', 'httpget', 'staging')
        client_mock.web_apps.list_function_keys.assert_not_called()



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from codecs import open
from setuptools import setup

VERSION = "0.3.0"

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Programming Language :: Python :: 3.13',
    'Programming Language :: Python :: 3.14',
    'License :: OSI Approved :: MIT License',
]

DEPENDENCIES = [
    'ConfigArgParse>=0.12.0',
    'jmespath',
    'vcrpy>=1.10.3',
    'pytest'
]

with open('README.rst', 'r', encoding='utf-8') as f:
    README = f.read()
with open('HISTORY.rst', 'r', encoding='utf-8') as f:
    HISTORY = f.read()

setup(
    name='azure-cli-testsdk',
    version=VERSION,
    description='Microsoft Azure Command-Line Tools Test SDK Module',
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    author='Microsoft Corporation',
    author_email='azpycli@microsoft.com',
    url='https://github.com/Azure/azure-cli',
    zip_safe=False,
    classifiers=CLASSIFIERS,
    packages=[
        'azure.cli.testsdk',
        'azure.cli.testsdk.scenario_tests',
    ],
    install_requires=DEPENDENCIES
)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
import time

from azure.cli.testsdk.scenario_tests import AllowLargeResponse
from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer
from azure.cli.testsdk.decorators import serial_test


class CognitiveServicesModelTests(ScenarioTest):
    @serial_test()
    @AllowLargeResponse()
    @ResourceGroupPreparer()
    def test_cognitiveservices_model(self, resource_group):
        sname = self.create_random_name(prefix='cs_cli_test_', length=16)
        
        self.kwargs.update({
            'sname': sname,
            'kind': 'OpenAI',
            'sku': 'S0',
            'location': 'WESTEUROPE'
        })

        # test to create cognitive services account
        self.cmd('az cognitiveservices account create -n {sname} -g {rg} --kind {kind} --sku {sku} -l {location} --yes',
                 checks=[self.check('name', '{sname}'),
                         self.check('location', '{location}'),
                         self.check('sku.name', '{sku}'),
                         self.check('properties.provisioningState', 'Succeeded')])

        models = self.cmd('az cognitiveservices account list-models -n {sname} -g {rg}').get_output_in_json()

        self.assertTrue(len(models) > 0)

        # delete the cognitive services account
        ret = self.cmd('az cognitiveservices account delete -n {sname} -g {rg}')
        self.assertEqual(ret.exit_code, 0)



if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


AUTOSCALE_TIMEZONES = ["Dateline Standard Time",
                       "UTC-11",
                       "Hawaiian Standard Time",
                       "Alaskan Standard Time",
                       "Pacific Standard Time (Mexico)",
                       "Pacific Standard Time",
                       "US Mountain Standard Time",
                       "Mountain Standard Time (Mexico)",
                       "Mountain Standard Time",
                       "Central America Standard Time",
                       "Central Standard Time",
                       "Central Standard Time (Mexico)",
                       "Canada Central Standard Time",
                       "SA Pacific Standard Time",
                       "Eastern Standard Time",
                       "USEasternStandardTime",
                       "Venezuela Standard Time",
                       "Paraguay Standard Time",
                       "Atlantic Standard Time",
                       "Central Brazilian Standard Time",
                       "SA Western Standard Time",
                       "Pacific SA Standard Time",
                       "Newfoundland Standard Time",
                       "E. South America Standard Time",
                       "Argentina Standard Time",
                       "SA Eastern Standard Time",
                       "Greenland Standard Time",
                       "Montevideo Standard Time",
                       "Bahia Standard Time",
                       "UTC-02",
                       "Mid-Atlantic Standard Time",
                       "Azores Standard Time",
                       "Cape Verde Standard Time",
                       "Morocco Standard Time",
                       "UTC",
                       "GMT Standard Time",
                       "Greenwich Standard Time",
                       "W. Europe Standard Time",
                       "Central Europe Standard Time",
                       "Romance Standard Time",
                       "Central European Standard Time",
                       "W. Central Africa Standard Time",
                       "Namibia Standard Time",
                       "Jordan Standard Time",
                       "GTB Standard Time",
                       "Middle East Standard Time",
                       "Egypt Standard Time",
                       "Syria Standard Time",
                       "E. Europe Standard Time",
                       "South Africa Standard Time",
                       "FLE Standard Time",
                       "Israel Standard Time",
                       "Kaliningrad Standard Time",
                       "Libya Standard Time",
                       "Arabic Standard Time",
                       "Turkey Standard Time",
                       "Arab Standard Time",
                       "Belarus Standard Time",
                       "Russian Standard Time",
                       "E. Africa Standard Time",
                       "Iran Standard Time",
                       "Arabian Standard Time",
                       "Azerbaijan Standard Time",
                       "Russia Time Zone 3",
                       "Mauritius Standard Time",
                       "Georgian Standard Time",
                       "Caucasus Standard Time",
                       "Afghanistan Standard Time",
                       "West Asia Standard Time",
                       "Ekaterinburg Standard Time",
                       "Pakistan Standard Time",
                       "India Standard Time",
                       "Sri Lanka Standard Time",
                       "Nepal Standard Time",
                       "Central Asia Standard Time",
                       "Bangladesh Standard Time",
                       "N. Central Asia Standard Time",
                       "Myanmar Standard Time",
                       "SE Asia Standard Time",
                       "North Asia Standard Time",
                       "China Standard Time",
                       "North Asia East Standard Time",
                       "Singapore Standard Time",
                       "W. Australia Standard Time",
                       "Taipei Standard Time",
                       "Ulaanbaatar Standard Time",
                       "Tokyo Standard Time",
                       "Korea Standard Time",
                       "Yakutsk Standard Time",
                       "Cen. Australia Standard Time",
                       "AUS Central Standard Time",
                       "E. Australia Standard Time",
                       "AUS Eastern Standard Time",
                       "West Pacific Standard Time",
                       "Tasmania Standard Time",
                       "Vladivostok Standard Time",
                       "Russia Time Zone 10",
                       "Magadan Standard Time",
                       "Central Pacific Standard Time",
                       "Russia Time Zone 11",
                       "New Zealand Standard Time",
                       "UTC+12",
                       "Fiji Standard Time",
                       "Kamchatka Standard Time",
                       "Tonga Standard Time",
                       "Samoa Standard Time",
                       "Line Islands Standard Time"
                       ]


class NodeRoleType:  # pylint: disable=too-few-public-methods
    HEADNODE = "HeadNodeRole"
    WORKERNODE = "WorkerNodeRole"
    ZOOKEEPERNODE = "ZookeeperRole"
    EDGENODE = "EdgeNodeRole"
    HIBNODE = "HIBRole"
    KAFKAMANAGEMENTNODE = "KafkaManagementNodeRole"


def get_key_for_storage_account(cmd, storage_account):  # pylint: disable=unused-argument
    from ._client_factory import cf_storage
    from azure.mgmt.core.tools import parse_resource_id, is_valid_resource_id
    from knack.util import CLIError

    storage_account_key = None
    if is_valid_resource_id(storage_account):
        parsed_storage_account = parse_resource_id(storage_account)
        resource_group_name = parsed_storage_account['resource_group']
        storage_account_name = parsed_storage_account['resource_name']

        storage_client = cf_storage(cmd.cli_ctx)
        keys = storage_client.storage_accounts.list_keys(resource_group_name, storage_account_name)
        storage_account_key = keys.keys_property[0].value  # pylint: disable=no-member
    elif storage_account:
        raise CLIError('Failed to get access key for storage account: {}'.format(storage_account))
    return storage_account_key


def get_storage_account_endpoint(cmd, storage_account, is_wasb):
    from ._client_factory import cf_storage
    from azure.mgmt.core.tools import parse_resource_id, is_valid_resource_id
    host = None
    if is_valid_resource_id(storage_account):
        parsed_storage_account = parse_resource_id(storage_account)
        resource_group_name = parsed_storage_account['resource_group']
        storage_account_name = parsed_storage_account['resource_name']

        storage_client = cf_storage(cmd.cli_ctx)
        storage_account = storage_client.storage_accounts.get_properties(
            resource_group_name=resource_group_name,
            account_name=storage_account_name)

        def extract_endpoint(storage_account, is_wasb):
            if not storage_account:
                return None
            return storage_account.primary_endpoints.dfs if not is_wasb else storage_account.primary_endpoints.blob

        def extract_host(uri):
            import re
            return uri and re.search('//(.*)/', uri).groups()[0]

        host = extract_host(extract_endpoint(storage_account, is_wasb))
    return host


def is_wasb_storage_account(cmd, storage_account):
    from ._client_factory import cf_storage
    from azure.mgmt.core.tools import parse_resource_id, is_valid_resource_id
    is_wasb = True
    if is_valid_resource_id(storage_account):
        parsed_storage_account = parse_resource_id(storage_account)
        resource_group_name = parsed_storage_account['resource_group']
        storage_account_name = parsed_storage_account['resource_name']
        storage_client = cf_storage(cmd.cli_ctx)
        properties = storage_client.storage_accounts.get_properties(resource_group_name, storage_account_name)
        is_wasb = not properties.is_hns_enabled
    return is_wasb


def build_identities_info(identities):
    from azure.mgmt.hdinsight.models import ClusterIdentity, ResourceIdentityType
    identity = None
    if identities:
        identity_type = ResourceIdentityType.user_assigned
        identity = ClusterIdentity(type=identity_type)
        identity.user_assigned_identities = {e: {} for e in identities}

    return identity


def build_update_identities_info(assign_identity_type, identities):
    from azure.mgmt.hdinsight.models import ClusterIdentity
    identity = ClusterIdentity(type=assign_identity_type)
    if identities:
        identity.user_assigned_identities = {str(e): {} for sublist in identities for e in sublist}
    return identity


def build_virtual_network_profile(subnet):
    from azure.mgmt.core.tools import resource_id, parse_resource_id, is_valid_resource_id
    from azure.mgmt.hdinsight.models import VirtualNetworkProfile
    from knack.util import CLIError

    vnet_profile = None
    if is_valid_resource_id(subnet):
        parsed_subnet_id = parse_resource_id(subnet)
        subscription_name = parsed_subnet_id['subscription']
        resource_group_name = parsed_subnet_id['resource_group']
        vnet_namespace = parsed_subnet_id['namespace']
        vnet_type = parsed_subnet_id['type']
        vnet_name = parsed_subnet_id['name']
        vnet_id = resource_id(
            subscription=subscription_name,
            resource_group=resource_group_name,
            namespace=vnet_namespace,
            type=vnet_type,
            name=vnet_name)
        vnet_profile = VirtualNetworkProfile(id=vnet_id, subnet=subnet)
    elif subnet:
        raise CLIError('Invalid subnet: {}'.format(subnet))
    return vnet_profile


def parse_domain_name(domain):
    from azure.mgmt.core.tools import parse_resource_id, is_valid_resource_id
    domain_name = None
    if is_valid_resource_id(domain):
        parsed_domain_id = parse_resource_id(domain)
        domain_name = parsed_domain_id['resource_name']
    return domain_name


# Validate ESP cluster creation required parameters
def validate_esp_cluster_create_params(esp,
                                       cluster_name,
                                       resource_group_name,
                                       cluster_type,
                                       subnet,
                                       domain,
                                       cluster_admin_account,
                                       assign_identity,
                                       ldaps_urls,
                                       cluster_admin_password,
                                       cluster_users_group_dns):
    from knack.util import CLIError
    if esp:
        missing_params = []
        if not cluster_name:
            missing_params.append("--name/-n")
        if not resource_group_name:
            missing_params.append("--resource-group/-g")
        if not cluster_type:
            missing_params.append("--type/-t")
        if not subnet:
            missing_params.append("--subnet")
        if not domain:
            missing_params.append("--domain")
        if not cluster_admin_account:
            missing_params.append("--cluster-admin-account")
        if not cluster_users_group_dns:
            missing_params.append("--cluster-users-group-dns")
        if not assign_identity:
            missing_params.append("--assign-identity")

        if missing_params:
            raise CLIError('the following params are required  '
                           'when --esp is specified: {}'.format(', '.join(missing_params)))
    else:
        esp_params = []
        if domain:
            esp_params.append("--domain")
        if cluster_admin_account:
            esp_params.append("--cluster-admin_account")
        if ldaps_urls:
            esp_params.append("--ldaps-urls")
        if cluster_admin_password:
            esp_params.append("--cluster-admin-password")
        if cluster_users_group_dns:
            esp_params.append("--cluster-users-group-dns")

        if esp_params:
            raise CLIError('the following params are required only '
                           'when --esp is specified: {}'.format(', '.join(esp_params)))


def get_resource_id_by_name(cli_ctx, resource_type, resource_name):
    from ._client_factory import cf_resources
    from knack.util import CLIError

    client = cf_resources(cli_ctx)
    filter_str = "resourceType eq '{}' and name eq '{}'".format(resource_type, resource_name) if resource_type else None
    resources = list(client.resources.list(filter=filter_str))
    if not resources:
        raise CLIError('Fails to retrieve any resource by name {}'.format(resource_name))
    if len(resources) > 1:
        raise CLIError('Found more than one resources by name {}. '
                       'Please specify one of the following resource IDs explicitly:\n{}'
                       .format(resource_name, '\n'.join([resource.id for resource in resources])))
    return resources[0].id


def get_default_vm_sizes_configurations(cli_ctx, location):
    from ._client_factory import cf_hdinsight_locations
    locations_client = cf_hdinsight_locations(cli_ctx)
    billing_response_result = locations_client.list_billing_specs(location)
    default_vm_sizes_configurations = {}
    for vm_filter in billing_response_result.vm_size_filters:
        if vm_filter.filter_mode == "Default":
            for node_type in vm_filter.node_types:
                cluster_type_vm_sizes_dict = default_vm_sizes_configurations.get(node_type.upper())
                if not cluster_type_vm_sizes_dict:
                    cluster_type_vm_sizes_dict = {}
                for cluster_type in vm_filter.cluster_flavors:
                    for vm_size in vm_filter.vm_sizes:
                        cluster_type_vm_sizes_dict[cluster_type.upper()] = vm_size
                default_vm_sizes_configurations[node_type.upper()] = cluster_type_vm_sizes_dict
    return default_vm_sizes_configurations


def get_default_vm_size_per_node_cluster_type(cluster_type, node_type, default_vm_sizes_configurations):
    cluster_type_vm_size_dict = default_vm_sizes_configurations.get(node_type.upper())
    cluster_type = map_cluster_type(cluster_type)

    vm_size = None
    if cluster_type_vm_size_dict:
        vm_size = cluster_type_vm_size_dict.get(cluster_type.upper())
        if not vm_size:
            vm_size = cluster_type_vm_size_dict.get("*")
    return vm_size


def set_vm_size(cli_ctx, location, cluster_type, headnode_size, workernode_size):
    if not headnode_size or not workernode_size:
        default_vm_size_configurations = get_default_vm_sizes_configurations(cli_ctx, location)
        if not headnode_size:
            headnode_size = get_default_vm_size_per_node_cluster_type(cluster_type, NodeRoleType.HEADNODE,
                                                                      default_vm_size_configurations)
        if not workernode_size:
            workernode_size = get_default_vm_size_per_node_cluster_type(cluster_type, NodeRoleType.WORKERNODE,
                                                                        default_vm_size_configurations)

    return headnode_size, workernode_size


def map_cluster_type(cluster_type):
    # the cluster type is mlservice in the list billing spec response
    if cluster_type.lower() == 'mlservices' or cluster_type.lower() == 'rserver':
        cluster_type = 'mlservice'
    return cluster_type


def get_entra_user_info(cmd, entra_user_identity, entra_user_full_info, toJson=True):
    import json
    from ._client_factory import cf_graph
    from azure.cli.core.azclierror import ResourceNotFoundError, InvalidArgumentValueError, AzureResponseError

    def is_email(value):
        return "@" in value and "." in value

    def normalize_keys(d):
        return {k.lower(): v for k, v in d.items()}

    graph_client = cf_graph(cmd.cli_ctx)
    rest_auth_entra_users = []
    if entra_user_identity:
        user_data = []
        for item in entra_user_identity:
            if item:
                user_data.extend([s.strip() for s in item.split(',') if s.strip()])
        for data in user_data:
            try:
                user = graph_client.user_get(data)
                if user is None:
                    filter_expr = f"id eq '{data}'"
                    if is_email(data):
                        filter_expr = f"userPrincipalName eq '{data}' or mail eq '{data}'"
                    user = graph_client.user_list(filter=filter_expr)
                    user = user[0]
                if user is None:
                    raise ResourceNotFoundError(
                        error_msg=f'No Entra user found for input: "{data}"',
                        recommendation=[
                            'Verify the user exists in Microsoft Entra ID',
                            'Confirm the identifier (email, upn, or objectId) is correct',
                            'Try querying manually: az ad user show --id <identifier>'
                        ]
                    )
                rest_auth_entra_users.append({'ObjectId': user['id'],
                                              'DisplayName': user['displayName'],
                                              'Upn': user['userPrincipalName']})
            except ResourceNotFoundError:
                raise
            except Exception as ex:
                raise AzureResponseError(
                    error_msg=f'Error: {type(ex).__name__}: {str(ex)}',
                    recommendation=[
                        'Check network connectivity to Microsoft Graph API',
                        'Validate authentication permissions for directory queries',
                        'Retry operation with --debug to inspect detailed error',
                        'Alternatively, consider using --entra-user-full-info to provide user details directly',
                    ]
                )
    else:
        for user in entra_user_full_info:
            user_normalized = normalize_keys(user)
            allowed_keys = {'objectid', 'displayname', 'upn'}
            if invalid_keys := set(user_normalized.keys()) - allowed_keys:
                raise InvalidArgumentValueError(
                    error_msg=f'Invalid keys detected in user object: {", ".join(invalid_keys)}',
                    recommendation=[
                        'User objects must only contain: objectId, displayName, and upn',
                        f'Remove invalid keys: {", ".join(invalid_keys)}',
                        'Example valid format: [{"objectId": "...", "displayName": "...", "upn": "..."}]'
                    ]
                )
            rest_auth_entra_users.append({'ObjectId': user_normalized.get('objectid'),
                                          'DisplayName': user_normalized.get('displayname'),
                                          'Upn': user_normalized.get('upn')})
    return json.dumps(rest_auth_entra_users) if toJson else rest_auth_entra_users




# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

class AttestationConstants:
    MANUAL_POLICY_RG_ASSIGNMENT = "7a00c0f2284a40128f1d3d22"
    MANUAL_POLICY_RG_INITIATIVE_ASSIGNMENT = "e148d7cc6585407bb1638382"
    MANUAL_POLICY_RG_INITIATIVE_REFID = "cliTests:RGManualPolicy_1"

    MANUAL_POLICY_SUB_ASSIGNMENT = "e861d6390e834761982c1c25"
    MANUAL_POLICY_SUB_INITIATIVE_ASSIGNMENT = "e1932be59a1c4b97bde62716"
    MANUAL_POLICY_SUB_INITIATIVE_REFID = "cliTests:SubscriptionManualPolicy_1"

    MANUAL_POLICY_RESOURCE_ASSIGNMENT = "ccbf75fff357452e88ca5532"
    MANUAL_POLICY_RESOURCE_INITIATIVE_ASSIGNMENT = "1213836084664464920a1b65"
    MANUAL_POLICY_RESOURCE_INITIATIVE_REFID = "cliTests:ResourceManualPolicy_1"

    ATTESTATION_TEST_RESOURCE_NAME = "CLI-TEST-VN"
    ATTESTATION_TEST_RG_NAME = "cli_test_attestations"
    ATTESTATION_RESOURCE_TYPE = "/providers/microsoft.policyinsights/attestations"
